import logging
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
from scipy import interpolate, optimize
from scipy.stats import binom, chi2

from .models.constants import M76, NA

# plotting colors
NICE_BLUE = "#668DA5"
NICE_RED = "#B4584D"
NICE_GREEN = "#ABB1A2"
NICE_PINK = "#CCACAD"


log = logging.getLogger(__name__)


def emp_cdf(
    data: np.array,  # the data to make a cdf out of
    bins=100,  # either number of bins or list of bin edges
) -> Tuple[np.array, np.array]:
    """
    Parameters
    ----------
    data
        unbinned data
    bins
        number of bins or array-like of bin edges

    Create a binned empirical CDF given a dataset. Empirical CDF is evaluated at right bin edge; the value corresponds to
    the PDF integrated up to the right bin edge.
    """

    if isinstance(bins, int):
        binedges = np.linspace(np.nanmin(data), np.nanmax(data), bins)
    elif isinstance(bins, np.ndarray) or isinstance(bins, list):
        binedges = np.array(bins)
    else:
        raise TypeError(f"bins must be array-like or int, instead is {type(bins)}")

    h, b = np.histogram(data, bins)

    return np.cumsum(h) / np.sum(h), b


def percentile(
    data: np.array,  # the data to make a cdf out of
    percentiles: np.array,  # which percentiles to find; should be in [0, 1]
):
    """
    Returns the test statistic that defines the percentiles in `percentiles`
    for the given data
    """
    nevts = len(data)
    data = sorted(data)
    if isinstance(percentiles, float):
        percentiles = np.array([percentiles])
    results = np.zeros_like(percentiles)
    for i, p in enumerate(percentiles):
        p_idx = int(p * nevts)
        p_rem = p * nevts - p_idx
        results[i] = data[p_idx]
    return results


def binomial_unc_band(
    cdf: np.array,  # binned CDF
    nevts: int,  # number of events the CDF is based off of
    CL: float = 0.68,  # confidence level for band
):
    """
    Returns the confidence band for a given CDF by taking the confidence interval of a
    binomial distribution with N=nevts and P=the value of the CDF at each point
    """
    interval = binom.interval(CL, nevts, cdf)
    lo_binom_band = interval[0] / nevts
    hi_binom_band = interval[1] / nevts

    return lo_binom_band, hi_binom_band


def dkw_band(
    cdf: np.array,  # binned CDF
    nevts: int,  # number of events the CDF is based off of
    CL: float = 0.68,  # confidence level for band
) -> Tuple[np.array, np.array]:
    """
    Returns the confidence band for a given CDF following the DKW inequality
    https://lamastex.github.io/scalable-data-science/as/2019/jp/11.html
    """
    alpha = 1.0 - CL
    eps = np.sqrt(np.log(2 / alpha) / (2 * nevts))
    lo_band = np.maximum(cdf - eps, np.zeros_like(cdf))
    hi_band = np.minimum(cdf + eps, np.ones_like(cdf))
    return lo_band, hi_band


def test_statistic_asymptotic_limit(
    t_mus: np.array, mu: float, mu_0: float, sigma: float
) -> np.array:
    """
    In the asymptotic limit, the test statiistics become distribute according to a noncentral chi-squared distribution

    Parameters
    ----------
    t_mus
        Array of test statistic values at which to evaluate the distribution
    mu
        Strength parameter under test
    mu_0
        Strength parameter the data are distributed according to
    sigma
        Standard deviation obtained from covariance matrix of estimators for all parameters
    """

    non_centrality = (mu - mu_0) ** 2 / sigma**2

    if non_centrality == 0:
        return 1 / np.sqrt(t_mus * 2 * np.pi)
    else:
        return (1 / np.sqrt(t_mus * 2 * np.pi)) * (
            np.exp(-0.5 * (np.sqrt(t_mus) + np.sqrt(non_centrality)) ** 2)
            + (np.exp(-0.5 * (np.sqrt(t_mus) - np.sqrt(non_centrality)) ** 2))
        )


def ts_critical(
    ts: np.array,  # list of test statistics (output of Experiment.toy_ts)
    threshold: float = 0.9,  # critical threshold for test statistic
    confidence: float = 0.68,  # width of confidence interval on the CDF
    plot: bool = False,  # if True, save plots of CDF and PDF with critical bands
    bins=None,  # int or array, number of bins or list of bin edges for CDF
    step: float = 0.01,  # specify the (approximate) step size for the bins if list of bins is not passed
    plot_dir: str = "",  # directory where to save plots
    plot_title: str = "",
):
    """
    Returns the critical value of the test statistic for the specified threshold and the confidence interval on this
    critical value.
    Bins are only used for plotting purposes
    """
    nevts = len(ts)
    binom_interval = binom.interval(confidence, nevts, threshold)
    lo_binom_percentile = binom_interval[0] / nevts
    hi_binom_percentile = binom_interval[1] / nevts

    ts_crit, ts_lo, ts_hi = percentile(
        data=ts, percentiles=[threshold, lo_binom_percentile, hi_binom_percentile]
    )

    if plot:
        if isinstance(bins, int):
            bins = np.linspace(np.floor(np.nanmin(ts)), np.nanmax(ts), bins)
        elif not isinstance(bins, np.ndarray):
            bins = np.linspace(
                np.floor(np.nanmin(ts)), np.nanmax(ts), int(np.nanmax(ts) / step)
            )

        cdf, binedges = emp_cdf(
            ts, bins
        )  # note that the CDF is evaluated at the right bin edge

        lo_band, hi_band = binomial_unc_band(cdf, nevts=len(ts), CL=confidence)

        int_thresh = int(100 * threshold)
        int_conf = int(100 * confidence)

        fig, axs = plt.subplots(1, 2, layout="constrained", figsize=(11, 5))

        axs[0].plot(binedges[1::], cdf, c="C0", label="empirical CDF")
        axs[0].plot(
            binedges[1::],
            lo_band,
            c="C0",
            ls="--",
            label=f"CDF {int_conf}% CL interval",
        )
        axs[0].plot(
            binedges[1::],
            hi_band,
            c="C0",
            ls="--",
        )
        axs[0].plot(
            binedges,
            chi2.cdf(binedges, df=1),
            color="black",
            label=r"$\chi^2_{dof=1}$ CDF",
        )

        axs[0].axvline(
            ts_crit,
            color="g",
            alpha=0.75,
            label=f"{int_thresh}% CL, "
            + rf"$t_C = {{{ts_crit:0.2f}}}_{{-{ts_crit-ts_lo:0.2f}}}^{{+{ts_hi-ts_crit:0.2f}}}$",
        )
        axs[0].axvspan(ts_lo, ts_hi, alpha=0.25, color="g")
        axs[0].axhline(
            threshold,
            color="orange",
            alpha=0.75,
            label=rf"actual CL: ${100*threshold:0.1f} \pm {100*(hi_binom_percentile-lo_binom_percentile)/2.0:0.1f}$%",
        )
        axs[0].axhspan(
            lo_binom_percentile, hi_binom_percentile, color="orange", alpha=0.25
        )
        axs[0].set_xlabel(r"$t$")
        axs[0].set_ylabel(r"CDF$(t)$")
        axs[0].legend()
        axs[0].set_ylim([0, 1])
        axs[0].set_xlim([bins[0], None])
        axs[0].grid()

        # bin the PDF with slightly larger bins if the step size is too small (just for viewing)
        bincenters = (binedges[1:] + binedges[:-1]) / 2.0
        binsize = bincenters[1] - bincenters[0]  # assume uniformly spaced bins
        factor = 1
        while (len(ts) / (len(binedges) / factor)) < 10:
            factor *= 2

        pdf = np.histogram(ts, bins=binedges[::factor])[0] / len(ts)

        axs[1].stairs(pdf, binedges[::factor], fill=False, zorder=3)
        axs[1].plot(
            bincenters,
            chi2.pdf(bincenters, df=1) * binsize * factor,
            color="k",
            label=r"$\chi^2_{dof=1}$ PDF",
        )
        axs[1].axvline(
            chi2.ppf(threshold, df=1),
            color="black",
            linestyle="dashed",
            label=f"{int_thresh}% CL, "
            + r"$\chi^2_C = $"
            + f"{chi2.ppf(threshold, df=1):0.2f}",
        )

        axs[1].axvline(
            ts_crit,
            color="g",
            alpha=0.75,
            label=f"{int_thresh}% CL, "
            + rf"$t_C = {{{ts_crit:0.2f}}}_{{-{ts_crit-ts_lo:0.2f}}}^{{+{ts_hi-ts_crit:0.2f}}}$",
        )
        axs[1].axvspan(ts_lo, ts_hi, color="g", alpha=0.25)

        axs[1].set_xlabel(r"t")
        axs[1].set_ylabel(r"$P(t)$")
        axs[1].legend()
        axs[1].set_yscale("log")
        axs[1].set_ylim([1 / (10 * len(ts)), 1.0])
        axs[1].set_xlim([bins[0], None])
        axs[1].grid(zorder=0)

        plt.suptitle(plot_title)
        plt.savefig(plot_dir + f"ts_critical_{int_thresh}.pdf", dpi=300)

        return (
            (ts_crit, ts_lo, ts_hi),
            (threshold, lo_binom_percentile, hi_binom_percentile),
            fig,
        )

    return (ts_crit, ts_lo, ts_hi), (
        threshold,
        lo_binom_percentile,
        hi_binom_percentile,
    )


def p_value(ts: np.array, ts_exp: float):
    """
    Returns p-value and uncertainty of an experimental test statistic and distribution of toy test statistics.
    """
    tot = 1.0 * len(ts)
    hi = np.sum(ts > ts_exp)

    pval = hi / tot

    # binomial uncertainty at 1 sigma
    unc = 1.0 / np.sqrt(tot) * np.sqrt(pval * (1.0 - pval))

    return pval, unc


def toy_ts_critical_p_value(
    ts: np.array,  # list of test statistics (output of Experiment.toy_ts)
    ts_exp: float,  # value of the test-statistic from experiment data at this s value
    bins=100,  # int or array, number of bins or list of bin edges for plotting
    step: float = 0.01,  # specify the (approximate) step size for the bins if list of bins is not passed, for plotting
    plot: bool = False,  # if True, save plots of CDF and PDF with critical bands
    plot_dir: str = "",  # directory where to save plots
    plot_title: str = "",
):
    """
    Returns the p-value associated with the observed test statistic from an experiment by comparing with the PDF generated from toys
    """

    ts_sorted = np.sort(ts)
    p_value = len(ts_sorted[ts_sorted >= ts_exp]) / len(ts_sorted)

    if plot:
        fig = plt.figure(figsize=(11, 6))

        pdf, binedges = np.histogram(ts, bins=bins)

        plt.stairs(pdf, binedges, color="C0", label="PDF")

        plt.axvline(
            ts_exp, color="g", alpha=0.75, label=rf"p-value: ${p_value*100:0.1f}$%"
        )

        plt.xlabel(r"$t$")
        plt.ylabel(r"PDF$(t)$")
        plt.legend()
        plt.grid()

        plt.suptitle(plot_title)
        plt.savefig(plot_dir + "ts_critical_p_value.pdf", dpi=300)

        return p_value, fig

    return p_value


def get_p_values(toy_ts: np.array, ts_observed: np.array):
    """
    Parameters
    ----------
    toy_ts
        List of lists. Each list is a list of test statistics from the toys generated at that value
    ts_observed
        List. These are the observed values of the test statistic from the experiment for a given value

    Returns
    -------
    p_values
        A list of the p-values associated with the observed data
    """

    if len(toy_ts) != len(ts_observed):
        log.warning(
            "The number of scanned points for the toys is not equal to the number of observed test statistics"
        )

    p_values = []
    for i, ts_exp in enumerate(ts_observed):
        p_values.append(toy_ts_critical_p_value(toy_ts[i], ts_exp, plot=False))

    return p_values


def sensitivity(
    toy_ts_zero_signal: list,
    toy_ts: list,
    s_list: list,
    CL: float = 0.9,
    bins=1000,  # int or array, number of bins or list of bin edges for CDF
    step: float = 0.01,  # specify the (approximate) step size for the bins if list of bins is not passed
    plot: bool = False,  # if True, save plots of CDF and PDF with critical bands
    plot_dir: str = "",  # directory where to save plots
    plot_title: str = "",
    save: bool = False,
):
    """
    Parameters
    ----------
    toy_ts
        List of lists. Each list is a a list of test statistics for toys generated with non-zero signal tested against that non-zero s-value
    toy_ts_zero_signal
        List of lists. Each list test statistics generated with zero signal and tested against a non-zero signal hypothesis
    s_list
        the list of the s_values being tested against

    Returns
    -------
    p_values
        A list of the p-values associated with the observed data


    Notes
    -----
    This function loops over the S-grid tested. At each value of S, the median of the PDF of toy_ts_zero_signal (generated with 0 signal and tested against S)
    is projected onto the PDF toy_ts (generated at S and tested against S) and the p-value of the median is calculated.
    """

    if len(toy_ts) != len(toy_ts_zero_signal):
        raise ValueError("input arrays must match in length!")

    p_values_median = []
    p_values_hi = []
    p_values_lo = []

    gammas = (
        M76 * s_list / (np.log(2) * NA)
    )  # convert from reduced s-value to half-rate

    for i, ts in enumerate(toy_ts_zero_signal):
        ts = np.array(ts)

        median_ts = np.median(ts)
        upper_ts = np.quantile(ts, 0.5 + CL / 2, method="linear")
        lower_ts = np.quantile(ts, 0.5 - CL / 2, method="linear")

        sorted_toy_ts = np.sort(toy_ts[i])

        p_values_median.append(
            len(sorted_toy_ts[sorted_toy_ts >= median_ts]) / len(sorted_toy_ts)
        )
        p_values_hi.append(
            len(sorted_toy_ts[sorted_toy_ts >= upper_ts]) / len(sorted_toy_ts)
        )
        p_values_lo.append(
            len(sorted_toy_ts[sorted_toy_ts >= lower_ts]) / len(sorted_toy_ts)
        )

        if plot:
            fig = plt.figure(figsize=(18, 10))

            pdf_zero_sig, binedges_zero_sig = np.histogram(
                ts, bins=bins, range=(-2, 25)
            )
            plt.stairs(
                pdf_zero_sig,
                binedges_zero_sig,
                color=NICE_RED,
                label=rf"PDF for $\Gamma$'={s_list[i]:.2e}|$\Gamma$=0",
            )

            pdf, binedges = np.histogram(toy_ts[i], bins=bins, range=(-2, 25))

            plt.stairs(
                pdf,
                binedges,
                color=NICE_BLUE,
                label=rf"PDF for $\Gamma$'={s_list[i]:.2e}|$\Gamma$={s_list[i]:.2e}",
            )
            plt.axvline(
                median_ts,
                color=NICE_GREEN,
                alpha=0.95,
                label="Median 0-Signal Test-Statistic",
            )
            plt.axvline(
                upper_ts,
                color=NICE_GREEN,
                ls="--",
                alpha=0.95,
                label=f"Upper {CL*100:.2f}%",
            )
            plt.axvline(
                lower_ts,
                color=NICE_PINK,
                ls="--",
                alpha=0.95,
                label=f"Lower {CL*100:.2f}%",
            )

            plt.yscale("log")
            plt.xlabel(r"$t$")
            plt.ylabel(r"PDF$(t)$")
            plt.legend()
            plt.ylim([0.9, 1e5])
            plt.xlim([-1, 25])
            plt.grid()

            plt.suptitle(plot_title)
            if save:
                fig.savefig(
                    plot_dir + f"ts_critical_p_value_{s_list[i]:.4f}.jpg", dpi=300
                )
            else:
                plt.show()

    return p_values_median, p_values_hi, p_values_lo


def find_crossing(
    scanned_var: np.array,
    ts: np.array,
    t_crits: np.array,
    method: str = "l",
):
    """
    Parameters
    ----------
    scanned_var
        Values of the variable that is being scanned
    ts
        Values of the test statistic at the scanned values
    t_crits
        The critical value or values of the test statistic
    interpolation_mode
        The mode in which to interpolate the crossing between ts and t_crits. "i" for index before crossing, "l" for linear, and "q" for quadratic are supported

    Notes
    -----
    It is important that ts and t_crits are on the same grid!
    This can handle two crossings in case of discovery!
    """

    # First check if t_crits is a scalar, then turn it into an array of the length of ts
    if isinstance(t_crits, float) or isinstance(t_crits, int):
        t_crits = np.full(len(ts), t_crits)

    diff = t_crits - ts - 0  # now we can find some zeros!
    crossing_idxs = []
    for i in range(0, len(diff) - 1, 1):
        if diff[i] <= 0 < diff[i + 1]:
            crossing_idxs.append(i)
        if diff[i] > 0 >= diff[i + 1]:
            crossing_idxs.append(i)

    # If index before crossing interpolation do the following:
    if method == "i":
        crossing_points = scanned_var[crossing_idxs]

    # If linear mode, treat the t_crits as linear interpolation, as well as the ts
    if (method == "l") or (method == "complex"):
        crossing_points = []
        for i in crossing_idxs:
            alpha = (ts[i + 1] - ts[i]) / (scanned_var[i + 1] - scanned_var[i])
            beta = (t_crits[i + 1] - t_crits[i]) / (scanned_var[i + 1] - scanned_var[i])
            intersection = (ts[i] - t_crits[i] + (beta - alpha) * scanned_var[i]) / (
                beta - alpha
            )
            crossing_points.append(intersection)

    if method == "q":
        raise NotImplementedError("Not yet implemented! Sorry!")

    if method == "complex":  # use the first pass guesses and get crazy with it
        f = interpolate.interp1d(scanned_var, diff)

        if len(crossing_points) >= 1:
            new_crosses = []
            for cross in crossing_points:
                # walk  forwards to find a change in sign
                new_sign = -1 * np.sign(f(np.amax([cross - 0.03, scanned_var[0]])))
                upper_limit = None
                for s in scanned_var[
                    scanned_var > np.amax([cross - 0.03, scanned_var[0]])
                ]:
                    if np.sign(f(s)) == new_sign:
                        upper_limit = s
                if upper_limit is None:
                    new_crosses.append(np.nan)

                sol = optimize.root_scalar(
                    f,
                    bracket=[np.amax([cross - 0.03, scanned_var[0]]), upper_limit],
                    method="brentq",
                )
                new_crosses.append(sol.root)

            crossing_points = new_crosses

    return crossing_points



import matplotlib.pyplot as plt
import numba as nb
import numpy as np

import freqfit.models.constants as constants
from freqfit.utils import inspectparameters

nb_kwd = {
    "nopython": True,
    "parallel": False,
    "nogil": True,
    "cache": True,
    "fastmath": True,
}

# window
#     must be a 2D array of form e.g. `np.array([[0,1],[2,3]])`
#     where edges of window are monotonically increasing (this is not checked), in keV.
#     Default is typical analysis window.

# default analysis window and width
WINDOW = np.array(constants.WINDOW)

WINDOWSIZE = 0.0
for i in range(len(WINDOW)):
    WINDOWSIZE += WINDOW[i][1] - WINDOW[i][0]

FULLWINDOWSIZE = WINDOW[-1][1] - WINDOW[0][0]

SEED = 42  # set the default random seed


@nb.jit(**nb_kwd)
def nb_pdf(
    Es: np.array,
    a: float,
    BI: float,
    exp: float,
    check_window: bool = False,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    a   
        "normalized" slope between -1 and 1 inclusive
    BI
        rate of background in cts/exposure/energy (not used in pdf)
    exp
        exposure (not used in pdf)
    check_window
        whether to check if the passed Es fall inside of the window. Default is False and assumes that the passed Es
        all fall inside the window (for speed)
    """

    x0 = WINDOW[0][0]
    x1 = WINDOW[-1][1]

    mid = (x1 + x0) / 2.0

    # take normalized slope and convert to actual slope
    slope = a * (2.0 / (FULLWINDOWSIZE * FULLWINDOWSIZE))

    b = 1.0 / FULLWINDOWSIZE

    includedarea = 0.0
    for i in nb.prange(WINDOW.shape[0]):
        includedarea += (2.0 * (WINDOW[i][1] - WINDOW[i][0]) 
            * (slope * (-2.0 * mid + WINDOW[i][0] + WINDOW[i][1]) + 2 * b))

    totarea = (2.0 * (WINDOW[-1][1] - WINDOW[0][0]) 
        * (slope * (-2.0 * mid + WINDOW[0][0] + WINDOW[-1][1]) + 2 * b))

    amp = totarea / includedarea

    # Initialize and execute the for loop
    y = np.empty_like(Es, dtype=np.float64)
    for i in nb.prange(Es.shape[0]):
        y[i] = (slope * (Es[i] - mid) + b) * amp

    if check_window:
        for i in nb.prange(Es.shape[0]):
            inwindow = False
            for j in range(len(WINDOW)):
                if WINDOW[j][0] <= Es[i] <= WINDOW[j][1]:
                    inwindow = True
            if not inwindow:
                y[i] = 0.0

    return y


@nb.jit(**nb_kwd)
def nb_density(
    Es: np.array,
    a: float,
    BI: float,
    exp: float,
    check_window: bool = False,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    a   
        "normalized" slope between -1 and 1 inclusive
    BI
        rate of background in cts/exposure/energy
    exp
        exposure
    check_window
        whether to check if the passed Es fall inside of the window. Default is False and assumes that the passed Es
        all fall inside the window (for speed)
    """

    x0 = WINDOW[0][0]
    x1 = WINDOW[-1][1]

    mid = (x1 + x0) / 2.0

    # take normalized slope and convert to actual slope
    slope = a * (2.0 / (FULLWINDOWSIZE * FULLWINDOWSIZE))

    b = 1.0 / FULLWINDOWSIZE

    includedarea = 0.0
    for i in nb.prange(WINDOW.shape[0]):
        includedarea += (2.0 * (WINDOW[i][1] - WINDOW[i][0]) 
            * (slope * (-2.0 * mid + WINDOW[i][0] + WINDOW[i][1]) + 2 * b))

    totarea = (2.0 * (WINDOW[-1][1] - WINDOW[0][0]) 
        * (slope * (-2.0 * mid + WINDOW[0][0] + WINDOW[-1][1]) + 2 * b))

    amp = totarea / includedarea * BI * exp * WINDOWSIZE

    # Initialize and execute the for loop
    y = np.zeros_like(Es, dtype=np.float64)
    for i in nb.prange(Es.shape[0]):
        y[i] = (slope * (Es[i] - mid) + b) * amp

    if check_window:
        for i in nb.prange(Es.shape[0]):
            inwindow = False
            for j in range(len(WINDOW)):
                if WINDOW[j][0] <= Es[i] <= WINDOW[j][1]:
                    inwindow = True
            if not inwindow:
                y[i] = 0.0
    
    return WINDOWSIZE * BI * exp, y

@nb.jit(nopython=True, fastmath=True, cache=True, error_model="numpy")
def nb_extendedrvs(
    a: float,
    BI: float,
    exp: float,
    seed: int = SEED,
) -> (np.array, int):
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    a   
        "normalized" slope between -1 and 1 inclusive
    BI
        rate of background in cts/exposure/energy
    exp
        exposure
    """

    np.random.seed(seed)

    n_bkg = np.random.poisson(BI * exp * WINDOWSIZE)

    # preallocate for background draws
    Es = np.zeros(n_bkg, dtype=np.float64)

    x0 = WINDOW[0][0]
    x1 = WINDOW[-1][1]

    mid = (x1 + x0) / 2.0

    # take normalized slope and convert to actual slope
    slope = a * (2.0 / (FULLWINDOWSIZE * FULLWINDOWSIZE))
    
    # there's a precision issue with how I am drawing rvs when a is very small
    if abs(a) < 1E-12:
        slope = 0.0

    b = 1.0 / FULLWINDOWSIZE

    # find area of each section and percent of cdf 
    cumareas = np.zeros(len(WINDOW))
    percentages = np.zeros(len(WINDOW))
    areas = np.zeros(len(WINDOW))

    totarea = 0.0
    for i in nb.prange(WINDOW.shape[0]):
        area = 2.0 * (WINDOW[i][1] - WINDOW[i][0]) * (slope * (-2.0 * mid + WINDOW[i][0] + WINDOW[i][1]) + 2 * b)
        totarea += area
        cumareas[i] = totarea
        areas[i] = area

    for i in nb.prange(WINDOW.shape[0]):
        percentages[i] = cumareas[i] / totarea

    # figure out which window each count belongs to
    whichwindow = np.random.uniform(0.0, 1.0, n_bkg)
    numwindow = np.zeros(WINDOW.shape[0], dtype='i')
    for i in nb.prange(n_bkg):
        for j in nb.prange(WINDOW.shape[0]):
            if whichwindow[i] < percentages[j]:
                numwindow[j] += 1
                break

    m = slope
    # now draw the events for each window
    Es = np.zeros(n_bkg)
    k = 0
    for i in nb.prange(WINDOW.shape[0]):
        x0 = WINDOW[i][0]
        rvs = 0.5*areas[i]*np.random.uniform(0.0, 1.0, numwindow[i])
        for j in nb.prange(numwindow[i]):
            if m == 0.0:
                Es[k] = WINDOW[i][0] + rvs[j] / (2.0*b)
            else:
                Es[k] = ((b**2 + 2*b*m*(x0-mid) + m*(mid**2*m - 2*mid*m*x0 + m*x0**2 + rvs[j]))**0.5 - b + mid*m) / m
            
            k += 1

    return (Es, (n_bkg, 0))

class linear_bkg_gen:
    def __init__(self):
        self.parameters = inspectparameters(self.density)
        del self.parameters["check_window"] # used for plotting and stuff, screws up rvs since not present there
        pass

    def pdf(
        self,
        Es: np.array,
        a: float,
        BI: float,
        exp: float,
        check_window: bool = False,
    ) -> np.array:
        return nb_pdf(Es, a, BI, exp, check_window=check_window)

    # for iminuit ExtendedUnbinnedNLL
    def density(
        self,
        Es: np.array,
        a: float,
        BI: float,
        exp: float,
        check_window: bool = False,
    ) -> np.array:
        return nb_density(Es, a, BI, exp, check_window=check_window)

    def extendedrvs(
        self,
        a: float,
        BI: float,
        exp: float,
        seed: int = SEED,
    ) -> np.array:
        return nb_extendedrvs(a, BI, exp, seed=seed)

linear_bkg = linear_bkg_gen()



import importlib
import inspect
import logging

import numpy as np
import yaml

log = logging.getLogger(__name__)


# takes a model function and returns a dict of its parameters with their default value
def inspectparameters(
    func,
) -> dict:
    """
    Returns a `dict` of parameters that methods of this model take as keys. Values are default values of the
    parameters. Assumes the first argument of the model is `data` and not a model parameter, so this key is not
    returned.
    """
    # pulled some of this from `iminuit.util`
    try:
        signature = inspect.signature(func)
    except ValueError:  # raised when used on built-in function
        return {}

    r = {}
    for i, item in enumerate(signature.parameters.items()):
        if i == 0:
            continue

        name, par = item

        if (default := par.default) is par.empty:
            r[name] = "nodefaultvalue"
        else:
            r[name] = default

    return r

def load_config(
    file: str | dict,
) -> dict:
    """
    Loads a config file or dict and converts `str` for some fields to the appropriate objects.
    """

    # if it's not a dict, it might be a path to a file
    if not isinstance(file, dict):
        with open(file) as stream:
            # switch from safe_load to load in order to check for duplicate keys
            config = yaml.load(stream, Loader=UniqueKeyLoader)

    else:
        config = file

    # get list of models and cost functions to import
    models = set()
    costfunctions = set()
    if "datasets" not in config:
        msg = f"`datasets` not found in `{file if file is not dict else 'provided `dict`'}`"
        raise KeyError(msg)

    if "parameters" not in config:
        msg = f"`parameters` not found in `{file if file is not dict else 'provided `dict`'}`"
        raise KeyError(msg)

    for datasetname, dataset in config["datasets"].items():
        if "model" in dataset:
            models.add(dataset["model"])
        else:
            msg = f"dataset `{datasetname}` has no `model`"
            raise KeyError(msg)

        if "costfunction" in dataset:
            costfunctions.add(dataset["costfunction"])
        else:
            msg = f"dataset `{datasetname}` has no `costfunction`"
            raise KeyError(msg)

    if "constraints" in config:
        for constraintname, constraint in config["constraints"].items():
            if "parameters" not in constraint:
                msg = f"constraint `{constraintname}` has no `parameters`"
                raise KeyError(msg)
            else:
                # these need to be lists for other stuff
                if not isinstance(constraint["parameters"], list):
                    constraint["parameters"] = [constraint["parameters"]]
                if not isinstance(constraint["values"], list):
                    constraint["values"] = [constraint["values"]]
                if "uncertainty" in constraint and not isinstance(
                    constraint["uncertainty"], list
                ):
                    constraint["uncertainty"] = [constraint["uncertainty"]]
                if "covariance" in constraint and not isinstance(
                    constraint["covariance"], np.ndarray
                ):
                    constraint["covariance"] = np.asarray(constraint["covariance"])

    if "combined_datasets" in config:
        for groupname, group in config["combined_datasets"].items():
            if "model" in group:
                models.add(group["model"])
            else:
                msg = f"combined_datasets `{groupname}` has no `model`"
                raise KeyError(msg)

        if "costfunction" in group:
            costfunctions.add(group["costfunction"])
        else:
            msg = f"combined_datasets `{groupname}` has no `costfunction`"
            raise KeyError(msg)

    # this is specific to set up of 0vbb model
    for model in models:
        modelclassname = model.split(".")[-1]
        modelclass = getattr(importlib.import_module(model), modelclassname)

        for datasetname, dataset in config["datasets"].items():
            if dataset["model"] == model:
                dataset["model"] = modelclass

        if "combined_datasets" in config:
            for groupname, group in config["combined_datasets"].items():
                if group["model"] == model:
                    group["model"] = modelclass

    # specific to iminuit
    for costfunctionname in costfunctions:
        costfunction = getattr(
            importlib.import_module("iminuit.cost"), costfunctionname
        )

        for datasetname, dataset in config["datasets"].items():
            if dataset["costfunction"] == costfunctionname:
                dataset["costfunction"] = costfunction

        if "combined_datasets" in config:
            for groupname, group in config["combined_datasets"].items():
                if group["costfunction"] == costfunctionname:
                    group["costfunction"] = costfunction

    # convert any limits from string to python object
    for par, pardict in config["parameters"].items():
        if "limits" in pardict and type(pardict["limits"]) is str:
            pardict["limits"] = eval(pardict["limits"])

        if "physical_limits" in pardict and type(pardict["physical_limits"]) is str:
            pardict["physical_limits"] = eval(pardict["physical_limits"])

    if "options" in config:
        for option, optionval in config["options"].items():
            if optionval in ["none", "None"]:
                config["options"][option] = None

    return config


def grab_results(
    minuit,
    use_grid_rounding: bool = False,
    grid_rounding_num_decimals: dict = {},  # noqa: B006
) -> dict:
    # I checked whether we need to shallow/deep copy these and it seems like we do not

    toreturn = {}
    toreturn["errors"] = minuit.errors.to_dict()  # returns dict
    toreturn["fixed"] = minuit.fixed.to_dict()  # returns dict
    toreturn["fval"] = minuit.fval  # returns float
    toreturn["nfit"] = minuit.nfit  # returns int
    toreturn["npar"] = minuit.npar  # returns int
    toreturn["parameters"] = minuit.parameters  # returns tuple of str
    toreturn["tol"] = minuit.tol  # returns float
    toreturn["valid"] = minuit.valid  # returns bool
    toreturn["values"] = minuit.values.to_dict()  # returns dict

    if use_grid_rounding:
        toreturn["values"] = {
            key: np.around(value, grid_rounding_num_decimals[key])
            for key, value in minuit.values.to_dict().items()
        }  # returns dict
        toreturn["fval"] = minuit._fcn(
            toreturn["values"].values()
        )  # overwrite the fval with the truncated params

    return toreturn


# use this YAML loader to detect duplicate keys in a config file
# https://stackoverflow.com/a/76090386
class UniqueKeyLoader(yaml.SafeLoader):
    def construct_mapping(self, node, deep=False):
        mapping = set()
        for key_node, value_node in node.value:
            each_key = self.construct_object(key_node, deep=deep)
            if each_key in mapping:
                raise ValueError(
                    f"Duplicate Key: {each_key!r} is found in YAML File.\n"
                    f"Error File location: {key_node.end_mark}"
                )
            mapping.add(each_key)
        return super().construct_mapping(node, deep)


# negative of the exponent of scientific notation of a number
def negexpscinot(number):
    base10 = np.log10(abs(number))
    return int(-1 * np.floor(base10))



"""
This is the script that creates the s-grid for toys to be run on, used for job submission
"""
# import os module
import os

# turn off file locking
os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"

# import python modules
import argparse

import h5py
import numpy as np

from freqfit.limit import SetLimit
from freqfit.utils import load_config

# Setup the argument parser
__pars__ = argparse.ArgumentParser()

# include the arguments
__pars__.add_argument("-i", type=str, default="", help="path to the input yaml file")
__pars__.add_argument("-e", type=str, default="majorana", help="name of the experiment")
__pars__.add_argument(
    "-o",
    type=str,
    default="/home/sjborden/freqfit/gerda_toys",
    help="output path to a folder to dump all the files in",
)

# parse the arguments
__args__ = __pars__.parse_args()

# interpret the arguments
# proc_count = __args__.c if __args__.c < mp.cpu_count() else mp.cpu_count()
proc_input = __args__.i
proc_output = __args__.o
proc_experiment = __args__.e


if __name__ == "__main__":
    print("Starting to profile the test statistic!", flush=True)  # noqa:T201
    # Create an s-grid to profile the test statistic on
    x = np.linspace(0.1e-9, 0.2, 100)[:]
    x = np.around(x, 3)
    x[
        0
    ] = 1e-9  # This is the first value that the Brazil toys will be scanned at in limit.py, need to ensure both evaluated at same point

    p = SetLimit(load_config(proc_input), out_path=proc_output, name=proc_experiment)

    p.set_var_to_profile("global_S")
    ts_fine = p.scan_ts(x)

    # Save to a file
    f = h5py.File(proc_output + f"/{proc_experiment}_ts_wilks_fine.h5", "a")
    dset = f.create_dataset("s", data=x)
    dset = f.create_dataset("t_s", data=ts_fine)
    f.close()

    # Create the text file used for toy submission
    with open(proc_output + "/s.txt", "a") as file_out:
        for s_value in x:
            file_out.write(f"{s_value}\n")

    print("exiting program!", flush=True)  # noqa:T201



import numba as nb
import numpy as np

import freqfit.models.constants as constants
from freqfit.utils import inspectparameters

nb_kwd = {
    "nopython": True,
    "parallel": False,
    "nogil": True,
    "cache": True,
    "fastmath": True,
    "inline": "always",
}

QBB = constants.QBB
N_A = constants.NA
M_A = constants.MDET

# default analysis window and width
# window
#     uniform background regions to pull from, must be a 2D array of form e.g. `np.array([[0,1],[2,3]])`
#     where edges of window are monotonically increasing (this is not checked), in keV.
#     Default is typical analysis window.
WINDOW = np.array(constants.WINDOW)

WINDOWSIZE = 0.0
for i in range(len(WINDOW)):
    WINDOWSIZE += WINDOW[i][1] - WINDOW[i][0]

SEED = 42  # set the default random seed


@nb.jit(**nb_kwd)
def nb_pdf(
    Es: np.array,
    S: float,
    BI: float,
    delta: float,
    sigma: float,
    eff: float,
    effunc: float,
    effuncscale: float,
    exp: float,
    check_window: bool = False,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    S
        The signal rate, in units of counts/(kg*yr)
    BI
        The background index rate, in counts/(kg*yr*keV)
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    eff
        The global signal efficiency, unitless
    effunc
        uncertainty on the efficiency
    effuncscale
        scaling parameter of the efficiency
    exp
        The exposure, in kg*yr
    check_window
        whether to check if the passed Es fall inside of the window. Default is False and assumes that the passed Es
        all fall inside the window (for speed)

    Notes
    -----
    This function computes the following:
    mu_S = (eff + effuncscale * effunc) * exp * S
    mu_B = exp * BI * windowsize
    pdf(E) = 1/(mu_S+mu_B) * [mu_S * norm(E_j, QBB - delta, sigma) + mu_B/windowsize]
    """

    # Precompute the signal and background counts
    # mu_S = np.log(2) * (N_A * S) * (eff + effuncscale * effunc) * exp / M_A
    mu_S = S * (eff + effuncscale * effunc) * exp
    mu_B = exp * BI * WINDOWSIZE

    # Precompute the prefactors so that way we save multiplications in the for loop
    B_amp = exp * BI
    S_amp = mu_S / (np.sqrt(2 * np.pi) * sigma)

    # Initialize and execute the for loop
    y = np.empty_like(Es, dtype=np.float64)
    for i in nb.prange(Es.shape[0]):
        y[i] = (1 / (mu_S + mu_B)) * (
            S_amp * np.exp(-((Es[i] - QBB + delta) ** 2) / (2 * sigma**2)) + B_amp
        )

    if check_window:
        for i in nb.prange(Es.shape[0]):
            inwindow = False
            for j in range(len(WINDOW)):
                if WINDOW[j][0] <= Es[i] <= WINDOW[j][1]:
                    inwindow = True
            if not inwindow:
                y[i] = 0.0

    return y


@nb.jit(**nb_kwd)
def nb_density(
    Es: np.array,
    S: float,
    BI: float,
    delta: float,
    sigma: float,
    eff: float,
    effunc: float,
    effuncscale: float,
    exp: float,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    S
        The signal rate, in units of counts/(kg*yr)
    BI
        The background index rate, in counts/(kg*yr*keV)
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    eff
        The global signal efficiency, unitless
    effunc
        uncertainty on the efficiency
    effuncscale
        scaling parameter of the efficiency
    exp
        The exposure, in kg*yr

    Notes
    -----
    This function computes the following, faster than without a numba wrapper:
    mu_S = (eff + effuncscale * effunc) * exp * S
    mu_B = exp * BI * windowsize
    CDF(E) = mu_S + mu_B
    pdf(E) =[mu_S * norm(E_j, QBB - delta, sigma) + mu_B/windowsize]
    """

    # Precompute the signal and background counts
    # mu_S = np.log(2) * (N_A * S) * (eff + effuncscale * effunc) * exp / M_A
    # S *= 0.01
    # BI *= 0.0001
    mu_S = S * (eff + effuncscale * effunc) * exp
    mu_B = exp * BI * WINDOWSIZE

    if sigma == 0:
        return np.inf, np.full_like(Es, np.inf, dtype=np.float64)

    # Precompute the prefactors so that way we save multiplications in the for loop
    B_amp = exp * BI
    S_amp = mu_S / (np.sqrt(2 * np.pi) * sigma)

    # Initialize and execute the for loop
    y = np.empty_like(Es, dtype=np.float64)
    for i in nb.prange(Es.shape[0]):
        y[i] = S_amp * np.exp(-((Es[i] - QBB + delta) ** 2) / (2 * sigma**2)) + B_amp

    return mu_S + mu_B, y


@nb.jit(**nb_kwd)
def nb_log_density(
    Es: np.array,
    S: float,
    BI: float,
    delta: float,
    sigma: float,
    eff: float,
    effunc: float,
    effuncscale: float,
    exp: float,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    S
        The signal rate, in units of counts/(kg*yr)
    BI
        The background index rate, in counts/(kg*yr*keV)
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    eff
        The global signal efficiency, unitless
    effunc
        uncertainty on the efficiency
    effuncscale
        scaling parameter of the efficiency
    exp
        The exposure, in kg*yr

    Notes
    -----
    This function computes the following, faster than without a numba wrapper:
    mu_S = (eff + effuncscale * effunc) * exp * S
    mu_B = exp * BI * windowsize
    CDF(E) = mu_S + mu_B
    pdf(E) =[mu_S * norm(E_j, QBB - delta, sigma) + mu_B/windowsize]
    """

    # Precompute the signal and background counts
    # mu_S = np.log(2) * (N_A * S) * (eff + effuncscale * effunc) * exp / M_A
    # S *= 0.01
    # BI *= 0.0001
    mu_S = S * (eff + effuncscale * effunc) * exp
    mu_B = exp * BI * WINDOWSIZE

    if sigma == 0:
        return np.inf, np.full_like(Es, np.inf, dtype=np.float64)

    if mu_S + mu_B < 0:
        return 0, np.full_like(Es, -np.inf)

    # Precompute the prefactors so that way we save multiplications in the for loop
    B_amp = exp * BI
    S_amp = mu_S / (np.sqrt(2 * np.pi) * sigma)

    # Initialize and execute the for loop
    y = np.empty_like(Es, dtype=np.float64)
    for i in nb.prange(Es.shape[0]):
        pdf = S_amp * np.exp(-((Es[i] - QBB + delta) ** 2) / (2 * sigma**2)) + B_amp

        if pdf <= 0:
            y[i] = -np.inf

        # Make an approximation based on machine precision, following J. Detwiler's suggestion
        u = (S_amp * np.exp(-((Es[i] - QBB + delta) ** 2) / (2 * sigma**2))) / B_amp

        if u <= 1e-8:
            y[i] = np.log(B_amp) + u
        else:
            y[i] = np.log(pdf)

    return mu_S + mu_B, y


@nb.jit(**nb_kwd)
def nb_density_gradient(
    Es: np.array,
    S: float,
    BI: float,
    delta: float,
    sigma: float,
    eff: float,
    effunc: float,
    effuncscale: float,
    exp: float,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    S
        The signal rate, in units of counts/(kg*yr)
    BI
        The background index rate, in counts/(kg*yr*keV)
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    eff
        The global signal efficiency, unitless
    effunc
        uncertainty on the efficiency
    effuncscale
        scaling parameter of the efficiency
    exp
        The exposure, in kg*yr

    Notes
    -----
    This function computes the gradient of the density function and returns a tuple where the first element is the gradient of the CDF, and the second element is the gradient of the PDF.
    The first element has shape (K,) where K is the number of parameters, and the second element has shape (K,N) where N is the length of Es.
    mu_S = S * exp * (eff + effuncscale * effunc)
    mu_B = exp * BI * windowsize
    pdf(E) = [mu_S * norm(E_j, QBB + delta, sigma) + mu_B/windowsize]
    cdf(E) = mu_S + mu_B
    """

    # mu_S = np.log(2) * (N_A * S) * eff * exp / M_A
    mu_S = S * exp * (eff + effuncscale * effunc)
    mu_B = exp * BI * WINDOWSIZE

    grad_CDF = np.array(
        [
            (eff + effuncscale * effunc) * exp,
            exp * WINDOWSIZE,
            0,
            0,
            S * exp,
            S * exp * effuncscale,
            S * exp * effunc,
            BI * WINDOWSIZE + S * (eff + effuncscale * effunc),
        ]
    )

    grad_PDF = np.zeros(shape=(8, len(Es)))
    if sigma == 0:  # give up
        for i in nb.prange(Es.shape[0]):
            grad_PDF[0][i] = np.inf
            grad_PDF[1][i] = exp
            grad_PDF[2][i] = np.inf
            grad_PDF[3][i] = np.inf
            grad_PDF[4][i] = np.inf
            grad_PDF[5][i] = np.inf
            grad_PDF[6][i] = np.inf
            grad_PDF[7][i] = np.inf

    else:
        for i in nb.prange(Es.shape[0]):
            # For readability, don't precompute anything and see how performance is impacted
            grad_PDF[0][i] = (
                np.exp(-((Es[i] - QBB + delta) ** 2) / (2 * sigma**2))
                * (eff + effuncscale * effunc)
                * exp
            ) / (np.sqrt(2 * np.pi) * sigma)
            grad_PDF[1][i] = exp
            grad_PDF[2][i] = (
                (
                    np.exp(-((Es[i] - QBB + delta) ** 2) / (2 * sigma**2))
                    * (eff + effuncscale * effunc)
                    * exp
                    * S
                )
                * (delta + Es[i] - QBB)
                / (np.sqrt(2 * np.pi) * sigma**3)
            )
            grad_PDF[3][i] = (
                (
                    np.exp(-((Es[i] - QBB + delta) ** 2) / (2 * sigma**2))
                    * (eff + effuncscale * effunc)
                    * exp
                    * S
                )
                * ((-delta - Es[i] + QBB - sigma) * (-delta - Es[i] + QBB + sigma))
                / (np.sqrt(2 * np.pi) * sigma**4)
            )
            grad_PDF[4][i] = (
                np.exp(-((Es[i] - QBB + delta) ** 2) / (2 * sigma**2)) * exp * S
            ) / (np.sqrt(2 * np.pi) * sigma)
            grad_PDF[5][i] = (
                np.exp(-((Es[i] - QBB + delta) ** 2) / (2 * sigma**2))
                * exp
                * effuncscale
                * S
            ) / (np.sqrt(2 * np.pi) * sigma)
            grad_PDF[6][i] = (
                np.exp(-((Es[i] - QBB + delta) ** 2) / (2 * sigma**2))
                * exp
                * effunc
                * S
            ) / (np.sqrt(2 * np.pi) * sigma)
            grad_PDF[7][i] = (
                np.exp(-((Es[i] - QBB + delta) ** 2) / (2 * sigma**2))
                * (eff + effuncscale * effunc)
                * S
            ) / (np.sqrt(2 * np.pi) * sigma) + BI

    return grad_CDF, grad_PDF


@nb.jit(**nb_kwd)
def nb_logpdf(
    Es: np.array,
    S: float,
    BI: float,
    delta: float,
    sigma: float,
    eff: float,
    effunc: float,
    effuncscale: float,
    exp: float,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    S
        The signal rate, in units of counts/(kg*yr)
    BI
        The background index rate, in counts/(kg*yr*keV)
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    eff
        The global signal efficiency, unitless
    effunc
        uncertainty on the efficiency
    effuncscale
        scaling parameter of the efficiency
    exp
        The exposure, in kg*yr

    Notes
    -----
    This function computes the following:
    mu_S = (eff + effuncscale * effunc) * exp * S
    mu_B = exp * BI * windowsize
    logpdf(E) = log(1/(mu_S+mu_B) * [mu_S * norm(E_j, QBB - delta, sigma) + mu_B/windowsize])
    """

    # Precompute the signal and background counts
    # mu_S = np.log(2) * (N_A * S) * eff * exp / M_A
    mu_S = S * (eff + effuncscale * effunc) * exp
    mu_B = exp * BI * WINDOWSIZE

    if sigma == 0:  # need this check for fitting
        return np.full_like(
            Es, np.log(exp * BI / (mu_S + mu_B))
        )  # TODO: make sure this simplification makes sense in the limit

    # Precompute the prefactors so that way we save multiplications in the for loop
    B_amp = exp * BI
    S_amp = mu_S / (np.sqrt(2 * np.pi) * sigma)

    # Initialize and execute the for loop
    y = np.empty_like(Es, dtype=np.float64)
    for i in nb.prange(Es.shape[0]):
        pdf = (1 / (mu_S + mu_B)) * (
            S_amp * np.exp(-((Es[i] - QBB + delta) ** 2) / (2 * sigma**2)) + B_amp
        )

        if pdf <= 0:
            y[i] = -np.inf
        else:
            y[i] = np.log(pdf)

    return y


@nb.jit(nopython=True, fastmath=True, cache=True, error_model="numpy")
def nb_rvs(
    n_sig: int,
    n_bkg: int,
    delta: float,
    sigma: float,
    seed: int = SEED,
) -> np.array:
    """
    Parameters
    ----------
    n_sig
        Number of signal events to pull from
    n_bkg
        Number of background events to pull from
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    seed
        specify a seed, otherwise uses default seed

    Notes
    -----
    This function pulls from a Gaussian for signal events and from a uniform distribution for background events
    in the provided windows, which may be discontinuous.
    """

    np.random.seed(seed)

    # Get energy of signal events from a Gaussian distribution
    # preallocate for background draws
    Es = np.append(np.random.normal(QBB - delta, sigma, size=n_sig), np.zeros(n_bkg))

    # Get background events from a uniform distribution
    bkg = np.random.uniform(0, 1, n_bkg)

    breaks = np.zeros(shape=(len(WINDOW), 2))
    for i in range(len(WINDOW)):
        thiswidth = WINDOW[i][1] - WINDOW[i][0]

        if i > 0:
            breaks[i][0] = breaks[i - 1][1]

        if i < len(WINDOW):
            breaks[i][1] = breaks[i][0] + thiswidth / WINDOWSIZE

        for j in range(len(bkg)):
            if breaks[i][0] <= bkg[j] <= breaks[i][1]:
                Es[n_sig + j] = (bkg[j] - breaks[i][0]) * thiswidth / (
                    breaks[i][1] - breaks[i][0]
                ) + WINDOW[i][0]

    return Es


@nb.jit(nopython=True, fastmath=True, cache=True, error_model="numpy")
def nb_extendedrvs(
    S: float,
    BI: float,
    delta: float,
    sigma: float,
    eff: float,
    effunc: float,
    effuncscale: float,
    exp: float,
    seed: int = SEED,
) -> np.array:
    """
    Parameters
    ----------
    S
        expected rate of signal events in events/(kg*yr)
    BI
        rate of background events in events/(kev*kg*yr)
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    eff
        The global signal efficiency, unitless
    effunc
        uncertainty on the efficiency
    effuncscale
        scaling parameter of the efficiency
    exp
        The exposure, in kg*yr
    seed
        specify a seed, otherwise uses default seed

    Notes
    -----
    This function pulls from a Gaussian for signal events and from a uniform distribution for background events
    in the provided windows, which may be discontinuous.
    """
    # S *= 0.01
    # BI *= 0.0001

    np.random.seed(seed)

    n_sig = np.random.poisson(S * (eff + effuncscale * effunc) * exp)
    n_bkg = np.random.poisson(BI * exp * WINDOWSIZE)

    # Get energy of signal events from a Gaussian distribution
    # preallocate for background draws
    Es = np.append(np.random.normal(QBB - delta, sigma, size=n_sig), np.zeros(n_bkg))

    # Get background events from a uniform distribution
    bkg = np.random.uniform(0, 1, n_bkg)

    breaks = np.zeros(shape=(len(WINDOW), 2))
    for i in range(len(WINDOW)):
        thiswidth = WINDOW[i][1] - WINDOW[i][0]

        if i > 0:
            breaks[i][0] = breaks[i - 1][1]

        if i < len(WINDOW):
            breaks[i][1] = breaks[i][0] + thiswidth / WINDOWSIZE

        for j in range(len(bkg)):
            if breaks[i][0] <= bkg[j] <= breaks[i][1]:
                Es[n_sig + j] = (bkg[j] - breaks[i][0]) * thiswidth / (
                    breaks[i][1] - breaks[i][0]
                ) + WINDOW[i][0]

    return Es, (n_sig, n_bkg)


class correlated_efficiency_0vbb_gen:
    def __init__(self):
        self.parameters = inspectparameters(self.density)
        pass

    def pdf(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        effunc: float,
        effuncscale: float,
        exp: float,
        check_window: bool = False,
    ) -> np.array:
        return nb_pdf(
            Es, S, BI, delta, sigma, eff, effunc, effuncscale, exp, check_window
        )

    def logpdf(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        effunc: float,
        effuncscale: float,
        exp: float,
    ) -> np.array:
        return nb_logpdf(Es, S, BI, delta, sigma, eff, effunc, effuncscale, exp)

    # for iminuit ExtendedUnbinnedNLL
    def density(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        effunc: float,
        effuncscale: float,
        exp: float,
    ) -> np.array:
        return nb_density(Es, S, BI, delta, sigma, eff, effunc, effuncscale, exp)

    # for iminuit ExtendedUnbinnedNLL
    def density_gradient(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        effunc: float,
        effuncscale: float,
        exp: float,
    ) -> np.array:
        return nb_density_gradient(
            Es, S, BI, delta, sigma, eff, effunc, effuncscale, exp
        )

    # for iminuit ExtendedUnbinnedNLL
    def log_density(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        effunc: float,
        effuncscale: float,
        exp: float,
    ) -> np.array:
        return nb_log_density(Es, S, BI, delta, sigma, eff, effunc, effuncscale, exp)

    # should we have an rvs method for drawing a random number of events?
    # `extendedrvs`
    # needs to use same parameters as the rest of the functions...
    def rvs(
        self,
        n_sig: int,
        n_bkg: int,
        delta: float,
        sigma: float,
        seed: int = SEED,
    ) -> np.array:
        return nb_rvs(n_sig, n_bkg, delta, sigma, seed=seed)

    def extendedrvs(
        self,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        effunc: float,
        effuncscale: float,
        exp: float,
        seed: int = SEED,
    ) -> np.array:
        return nb_extendedrvs(
            S, BI, delta, sigma, eff, effunc, effuncscale, exp, seed=seed
        )

    def plot(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        effunc: float,
        effuncscale: float,
        exp: float,
    ) -> None:
        y = nb_pdf(Es, S, BI, delta, sigma, eff, effunc, effuncscale, exp)

        import matplotlib.pyplot as plt

        plt.step(Es, y)
        plt.show()

    # function call needs to take the same parameters as the other function calls, in the same order repeated twice
    # this is intended only for empty datasets
    # returns `None` if we couldn't combine the datasets (a dataset was not empty)
    def combine(
        self,
        a_Es: np.array,
        a_S: float,
        a_BI: float,
        a_delta: float,
        a_sigma: float,
        a_eff: float,
        a_effunc: float,
        a_effuncscale: float,
        a_exp: float,
        b_Es: np.array,
        b_S: float,
        b_BI: float,
        b_delta: float,
        b_sigma: float,
        b_eff: float,
        b_effunc: float,
        b_effuncscale: float,
        b_exp: float,
    ) -> list | None:
        # datasets must be empty to be combined
        if len(a_Es) != 0 or len(b_Es) != 0:
            return None

        Es = np.array([])  # both of these datasets are empty
        S = 0.0  # this should be overwritten in the fit later
        BI = 0.0  # this should be overwritten in the fit later

        exp = a_exp + b_exp  # total exposure

        # exposure weighted fixed parameters (important to calculate correctly)
        sigma = (a_exp * a_sigma + b_exp * b_sigma) / exp
        eff = (a_exp * a_eff + b_exp * b_eff) / exp
        delta = (a_exp * a_delta + b_exp * b_delta) / exp

        # these are fully correlated in this model so the direct sum is appropriate
        # (maybe still appropriate even if not fully correlated?)
        effunc = (a_exp * a_effunc + b_exp * b_effunc) / exp

        effuncscale = 0.0  # this should be overwritten in the fit later

        return [Es, S, BI, delta, sigma, eff, effunc, effuncscale, exp]

    def can_combine(
        self,
        a_Es: np.array,
        a_S: float,
        a_BI: float,
        a_delta: float,
        a_sigma: float,
        a_eff: float,
        a_effunc: float,
        a_effuncscale: float,
        a_exp: float,
    ) -> bool:
        """
        This sets an arbitrary rule if this dataset can be combined with other datasets.
        In this case, if the dataset contains no data, then it can be combined, but more complex rules can be imposed.
        """
        if len(a_Es) == 0:
            return True
        else:
            return False

    def intial_guess(self, Es: np.array, exp_tot: float, eff_tot: float) -> tuple:
        """
        Give a better initial guess for the signal and background rate given an array of data
        The signal rate is estimated in a +/-5 keV window around Qbb, the BI is estimated from everything outside that window

        Parameters
        ----------
        Es
            A numpy array of observed energy data
        exp_tot
            The total exposure of the experiment
        eff_tot
            The total efficiency of the experiment
        """
        QBB_ROI_SIZE = [
            5,
            5,
        ]  # how many keV away from QBB in - and + directions we are defining the ROI
        BKG_WINDOW_SIZE = WINDOWSIZE - np.sum(
            QBB_ROI_SIZE
        )  # subtract off the keV we are counting as the signal region
        n_sig = 0
        n_bkg = 0
        for E in Es:
            if QBB - QBB_ROI_SIZE[0] <= E <= QBB + QBB_ROI_SIZE[1]:
                n_sig += 1
            else:
                n_bkg += 1

        # find the expected BI
        BI_guess = n_bkg / (BKG_WINDOW_SIZE * exp_tot)

        # Now find the expected signal rate
        n_sig -= (
            n_bkg * np.sum(QBB_ROI_SIZE) / BKG_WINDOW_SIZE
        )  # subtract off the expected number of BI counts in ROI

        s_guess = n_sig / (exp_tot * eff_tot)

        return s_guess, BI_guess


correlated_efficiency_0vbb = correlated_efficiency_0vbb_gen()



"""
A class that holds a dataset and its associated model and cost function,
"""
import logging

import numpy as np
from iminuit import cost

SEED = 42

log = logging.getLogger(__name__)


class Dataset:
    def __init__(
        self,
        data: np.array,
        model,
        model_parameters: dict[str, str],
        parameters: dict,
        costfunction: cost.Cost,
        name: str = "",
        try_combine: bool = False,
        combined_dataset: str = None,
        user_gradient: bool = False,
        use_log: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        data
            `list` or `np.array` of the unbinned data
        model
            model `class` to be passed to the cost function. The `model`
            class must have a method `model.density(data, a, b, c, ...)` where `data` takes a 1D `ndarray` corresponding
            to unbinned events, and `a, b, c, ...` are the parameters of the model, which may take any type and may
            include default values. `model` must return the form expected by `costfunction`. Those parameters which
            are to be fit by `iminuit` must take a single number. `model` must have a `parameters` attribute that is a
            `dict` with keys of the parameter names `a, b, c, ...` and values of the default values of the parameters
            or "nodefaultvalue" if no default is provided. (e.g. the return of the `utils.inspectparameters()` method).
        model_parameters
            `dict` that contains the mapping between the named model parameters `a, b, c, ...` and a string containing
            a different name for the parameter that will be passed to `iminuit` in place of the original name.
            This can be used to fit the same parameter across multiple datasets since `iminuit` references variables
            by name.
        parameters
            `dict` for the parameters of the fit that control how the fit is performed.

            - "includeinfit": True or False, corresponding to whether this parameter should be included in the `iminuit`
                fit or if its value should be fixed. This allows for passing variables of any type to the underlying
                `model`. If True, no "value" will be passed to `model` for this parameter and will instead be taken
                from `iminuit` during the minimization. If "False", then "value" is required if the parameter does not
                have a default value and will be passed to `model`. Note that it is still possible to include
                parameters in the `iminuit` minimization and set them to `fixed`, which can be desirable in some cases.

            - "value": If "includeinfit" is `False` and no default value for the parameter is specified in `model`, then
                this is required. This may be any type and will be passed to `model`. If "includeinfit" is `True`, this
                is not used. If "includeinfit" is `False` and a default value for the parameter exists, this will
                overwrite the default.

            - "limits": This allows for setting limits on the parameters for `iminuit`.
        costfunction
            an `iminuit` cost function. Currently, only `cost.ExtendedUnbinnedNLL` or `cost.UnbinnedNLL` are supported
            as cost functions.
        name
            `str` name for the `Dataset`
        user_gradient
            If true use the user's gradient in the costfunction
        use_log
            If true, use the log of the PDF instead of the PDF within the Iminuit cost function

        Notes
        -----
        Currently, only `cost.ExtendedUnbinnedNLL` or `cost.UnbinnedNLL` are supported as cost functions.
        """

        self.data = np.asarray(data)  # the data of this Dataset
        self.name = name  # name of this Dataset
        self.model = model  # model object for this Dataset
        self.model_parameters = (
            model_parameters  # model parameter name : parameter name
        )

        self._costfunctioncall = None  # function call for the costfunction
        self.costfunction = None  # iminuit cost function object

        self._parlist = (
            []
        )  # list that contains all of the parameters of the model for this Dataset in the correct order
        self._parlist_indices = (
            []
        )  # indices in self._parlist of the the parameters to be fit

        self.fitparameters = (
            {}
        )  # fit parameter names : index in self._parlist (same as in self._parlist_indices)
        self._toy_pars_to_vary = (
            {}
        )  # parameter to vary for toys : index in self._parlist

        self.try_combine = try_combine  # whether to attempt to combine this Dataset into a combined_dataset
        self.is_combined = (
            False  # whether this Dataset is combined into a combined_dataset
        )
        self.combined_dataset = (
            None  # name of the combined_dataset that this Dataset is part of
        )
        self.user_gradient = user_gradient
        self.use_log = use_log

        self._toy_data = None  # place to hold Toy data for this Dataset
        self._toy_num_drawn = None  # place to hold the number of signal and number of background events drawn for a toy
        self._toy_is_combined = (
            False  # if this Dataset is combined into a combined_dataset in the Toy
        )

        if self.try_combine:
            if combined_dataset is None:
                msg = f"`Dataset` `{self.name}` has `try_combine` `{self.try_combine}` but `combined_dataset` is {combined_dataset}"
                raise ValueError(msg)
            else:
                self.combined_dataset = combined_dataset

        # check that all passed parameters are valid
        for parameter in model_parameters:
            if parameter not in self.model.parameters:
                msg = f"`Dataset` `{self.name}`: model_parameter `{parameter}` not found in model `{self.model}`."
                raise KeyError(msg)

        # check that all required parameters were passed
        for parameter, defaultvalue in self.model.parameters.items():
            if (defaultvalue is str and defaultvalue == "nodefaultvalue") and (
                parameter not in model_parameters
            ):
                msg = f"`Dataset` `{self.name}`: required model parameter `{parameter}` not found in model_parameters"
                raise KeyError(msg)

        # make the cost function
        if (costfunction is cost.ExtendedUnbinnedNLL) or (
            costfunction is cost.UnbinnedNLL
        ):
            self._costfunctioncall = costfunction
            if user_gradient:
                self.costfunction = costfunction(
                    self.data, self.density, grad=self.density_gradient
                )
            elif use_log:
                self.costfunction = costfunction(
                    self.data,
                    self.log_density,
                    log=True,
                )
            else:
                self.costfunction = costfunction(self.data, self.density)
        else:
            msg = f"`Dataset` `{self.name}`: only `cost.ExtendedUnbinnedNLL` or `cost.UnbinnedNLL` are supported as \
                cost functions"
            raise NotImplementedError(msg)

        # now we make the parameters of the cost function
        # need to go in order of the model
        for i, (par, defaultvalue) in enumerate(self.model.parameters.items()):
            # if not passed, use default value (already checked that required parameters passed)
            if par not in model_parameters:
                self._parlist.append(defaultvalue)
                break  # no constraints on non-passed parameters

            # parameter was passed and should be included in the fit
            elif ("includeinfit" in parameters[model_parameters[par]]) and (
                parameters[model_parameters[par]]["includeinfit"]
            ):
                self.costfunction._parameters |= {
                    model_parameters[par]: parameters[model_parameters[par]]["limits"]
                    if "limits" in parameters[model_parameters[par]]
                    else None
                }

                if "value" not in parameters[model_parameters[par]]:
                    msg = (
                        f"`Dataset` `{self.name}`: value for parameter `{par}` is required for"
                        + f" model `{model}` parameter `{par}`"
                    )
                    raise KeyError(msg)
                self._parlist.append(parameters[model_parameters[par]]["value"])
                self._parlist_indices.append(i)
                self.fitparameters |= {model_parameters[par]: i}

            else:  # parameter was passed but should not be included in the fit
                if ("value" not in parameters[model_parameters[par]]) and (
                    defaultvalue == "nodefaultvalue"
                ):
                    msg = (
                        f"`Dataset` '{self.name}': value for parameter '{par}' is required for"
                        + f"model '{model}' parameter '{par}'"
                    )
                    raise KeyError(msg)
                self._parlist.append(parameters[model_parameters[par]]["value"])

            if ("vary_by_constraint" in parameters[model_parameters[par]]) and (
                parameters[model_parameters[par]]["vary_by_constraint"]
            ):
                self._toy_pars_to_vary[model_parameters[par]] = i
                msg = f"`Dataset` '{self.name}': adding parameter '{model_parameters[par]}' as parameter to vary for toys"
                logging.info(msg)

        return

    def density(
        self,
        data,
        *par,  # DO NOT DELETE THE * - NEEDED FOR IMINUIT
    ) -> np.array:
        # par should be 1D array like
        # assign the positional parameters to the correct places in the model parameter list
        for i in range(len(par)):
            self._parlist[self._parlist_indices[i]] = par[i]

        return self.model.density(data, *self._parlist)

    def log_density(
        self,
        data,
        *par,  # DO NOT DELETE THE * - NEEDED FOR IMINUIT
    ) -> np.array:
        # par should be 1D array like
        # assign the positional parameters to the correct places in the model parameter list
        for i in range(len(par)):
            self._parlist[self._parlist_indices[i]] = par[i]

        return self.model.log_density(data, *self._parlist)

    def density_gradient(
        self,
        data,
        *par,  # DO NOT DELETE THE * - NEEDED FOR IMINUIT
    ) -> np.array:
        """
        Parameters
        ----------
        data
            Unbinned data
        par
            Potentially a subset of the actual model density_gradient parameters, depending on the config
        """
        # par should be 1D array like
        # assign the positional parameters to the correct places in the model parameter list
        for i in range(len(par)):
            self._parlist[self._parlist_indices[i]] = par[i]

        grad_cdf, grad_pdf = self.model.density_gradient(data, *self._parlist)

        # Mask the return values according to what the actual cost function expects
        return grad_cdf[self._parlist_indices], *grad_pdf[[self._parlist_indices], :]

    def rvs(
        self,
        par,
        seed: int = SEED,
    ) -> np.array:
        # par should be 1D array like
        # assign the positional parameters to the correct places in the model parameter list
        for i in range(len(par)):
            self._parlist[self._parlist_indices[i]] = par[i]

        rvs, toy_num_drawn = self.model.extendedrvs(*self._parlist, seed=seed)
        return rvs, toy_num_drawn

    # generates toy data and sets some attributes
    def toy(
        self,
        par,
        seed: int = SEED,
    ) -> None:
        self.toy_reset()
        self._toy_data, self._toy_num_drawn = self.rvs(par, seed=seed)
        return

    # resets some toy attributes
    def toy_reset(
        self,
    ) -> None:
        self._toy_data = None
        self._toy_num_drawn = None
        self._toy_is_combined = False
        return


# method to combine datasets
def combine_datasets(
    datasets: list[Dataset, ...],
    model,
    model_parameters: dict[str, str],
    parameters: dict,
    costfunction: cost.Cost,
    name: str = "",
    use_toy_data: bool = False,
    user_gradient: bool = False,
    use_log: bool = False,
) -> Dataset:
    """
    Parameters
    ----------
    datasets
        tuple of the `Dataset` to combine. The `model` must have a `combine` method that specifies whether and how the
        datasets are to be combined.
    model
        model of the combined `Dataset` (see `Dataset`)
    model_parameters
        see `Dataset`
    parameters
        see `Dataset`
    costfunction
        see `Dataset`
    name
        name of the combined `Dataset`
    user_gradient
        If true, use the user gradient in the combined dataset costfunction
    use_log
        If true, use the log of the PDF in the combined dataset costfunction
    """

    if not isinstance(datasets, list) or not all(
        isinstance(ds, Dataset) for ds in datasets
    ):
        msg = "must be `list` of `Datasets` to be combined"
        raise TypeError(msg)

    # if len(datasets) == 1:
    #     msg = f"only one `Dataset` named `{datasets[0].name}` was provided to combine in group `{name}`"
    #     logging.warning(msg)

    included_datasets = []
    combination = None
    first = True
    for ds in datasets:
        if ds.model != model:
            msg = f"`Dataset` {ds.name} has model `{ds.model}` != `{model}` - must be the same to combine"
            raise NotImplementedError(msg)

        if ds._costfunctioncall != costfunction:
            msg = f"`Dataset` {ds.name} has costfunction `{ds._costfunctioncall}` != `{costfunction}` - must be the same to combine"
            raise NotImplementedError(msg)

        # whether to use the real data or the toy data
        thisdata = ds.data
        if use_toy_data:
            if ds._toy_data is None:
                msg = f"`Dataset` '{ds.name}' has no toy data - you need to call `Dataset.rvs()` to make this."
                logging.error(msg)
                raise ValueError(msg)
            thisdata = ds._toy_data

        # if first dataset, the combination is just the stuff needed to recreate it
        # TODO: check if it is allowed to be combined
        if first and (ds.model.can_combine(thisdata, *ds._parlist)):
            combination = [thisdata, *ds._parlist]
            included_datasets.append(ds.name)
            msg = f"added `Dataset` '{ds.name}' to combined dataset '{name}'"
            logging.info(msg)
            first = False
            if use_toy_data:
                ds._toy_is_combined = True
            else:
                ds.is_combined = True
        elif ds.model.can_combine(thisdata, *ds._parlist):
            result = model.combine(*combination, thisdata, *ds._parlist)
            if result is not None:  # if None, we cannot combine them
                combination = result
                included_datasets.append(ds.name)
                msg = f"added `Dataset` '{ds.name}' to combined dataset '{name}'"
                logging.info(msg)
                if use_toy_data:
                    ds._toy_is_combined = True
                else:
                    ds.is_combined = True
            else:
                msg = f"not able to combine `Dataset` '{ds.name}' with combined dataset '{name}', will be kept separate"
                logging.info(msg)

    # return the combined dataset if we combined some datasets
    combined_dataset = None
    if len(included_datasets) > 0:
        # now to set the parameters based on the combined results
        simplified_parameters = {}
        # model.parameters contains all parameters, including default valued ones
        for i, par in enumerate(model.parameters.keys()):
            # let Dataset handle errors if parameters are missing
            if par in model_parameters:
                parname = model_parameters[par]
                # this is a reference not a copy!
                simplified_parameters[parname] = parameters[parname]
                # hence why it is important to check whether we should overwrite the value
                if (
                    "value_from_combine" in parameters[parname]
                    and parameters[parname]["value_from_combine"]
                ):
                    # i+1 because data occupies index 0
                    simplified_parameters[parname]["value"] = combination[i + 1]

        data = combination[0]

        combined_dataset = Dataset(
            data,
            model,
            model_parameters,
            simplified_parameters,
            costfunction,
            name,
            user_gradient=user_gradient,
            use_log=use_log,
        )

    return combined_dataset, included_datasets



"""
Models for unbinned frequentist fitting. These models are classes and come with methods to compute the pdf and logpdf.
"""

from freqfit.models.box_model_0vbb import box_model_0vbb
from freqfit.models.correlated_efficiency_0vbb import correlated_efficiency_0vbb
from freqfit.models.correlated_efficiency_0vbb_correlate_delta import (
    correlated_efficiency_0vbb_correlate_delta,
)
from freqfit.models.correlated_efficiency_0vbb_exponential_background import (
    correlated_efficiency_0vbb_exponential_background,
)
from freqfit.models.correlated_efficiency_0vbb_linear_background import (
    correlated_efficiency_0vbb_linear_background,
)
from freqfit.models.correlated_efficiency_NME_0vbb import (
    correlated_efficiency_NME_0vbb,
)
from freqfit.models.gaussian_on_uniform import gaussian_on_uniform
from freqfit.models.linear_bkg import linear_bkg
from freqfit.models.mjd_0vbb import mjd_0vbb
from freqfit.models.mjd_NME_0vbb import mjd_NME_0vbb
from freqfit.models.onebin_poisson import onebin_poisson
from freqfit.models.truncated_correlated_efficiency_0vbb import (
    truncated_correlated_efficiency_0vbb,
)

__all__ = [
    "gaussian_on_uniform",
    "correlated_efficiency_0vbb",
    "mjd_0vbb",
    "mjd_NME_0vbb",
    "onebin_poisson",
    "linear_bkg",
    "correlated_efficiency_0vbb_exponential_background",
    "correlated_efficiency_0vbb_linear_background",
    "box_model_0vbb",
    "truncated_correlated_efficiency_0vbb",
    "correlated_efficiency_NME_0vbb",
    "correlated_efficiency_0vbb_correlate_delta",
]



"""
A class that holds a combination of `Dataset` and `NormalConstraint`.
"""
import logging
from copy import deepcopy

import numpy as np
from iminuit import cost

from freqfit.dataset import Dataset, combine_datasets

SEED = 42

log = logging.getLogger(__name__)


class Superset:
    def __init__(
        self,
        datasets: dict,
        parameters: dict,
        constraints: dict = None,
        combined_datasets: dict = None,
        name: str = "",
        try_to_combine_datasets: bool = False,
        user_gradient: bool = False,
        use_log: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        datasets
            `dict`
        parameters
            `dict`
        constraints
            `dict`
        user_gradient
            If true, use the user's gradient in dataset costfunction
        use_log
            Use the logpdf in the cost function if true
        """

        self.name = name  # name of the Superset

        # parameter dictionary that contains all parameters used in any Datasets and combined_datasets. Not necessarily fit parameters.
        self.parameters = parameters

        self.datasets = {}  # dataset name: Dataset
        self.combined_datasets = {}  # combined_dataset name : Dataset
        self._combined_datasets_config = (
            combined_datasets  # config for combined_datasets for use in Toy
        )
        self.included_in_combined_datasets = (
            {}
        )  # combined_dataset name : list of dataset names contained

        self.costfunction = None  # iminuit costfunction object that will contain the Dataset costfunctions and NormalConstraint
        self.fitparameters = (
            None  # reference to self.costfunction._parameters, parameters of the fit
        )

        self.constraints = None  # the constraint costfunction
        self.constraints_parameters = {}  # parameter : index in values and covariance
        self.constraints_values = None
        self.constraints_covariance = None

        self._toy_pars_to_vary = []  # list of parameters to vary
        self._toy_parameters = (
            {}
        )  # a copy of self.parameters that we can mutate without care

        if try_to_combine_datasets:
            msg = "option 'try_to_combine_datasets' set to True - will attempt to combine `Dataset` where indicated"
            logging.info(msg)

            if combined_datasets is None or len(combined_datasets) == 0:
                msg = "option 'try_to_combine_datasets' set to True but `combined_datasets` is missing or empty!"
                logging.error(msg)

        alldspars = set()
        # create the Datasets
        for dsname in datasets.keys():
            try_combine = False
            combined_dataset = None
            if try_to_combine_datasets:
                try_combine = (
                    datasets[dsname]["try_to_combine"]
                    if "try_to_combine" in datasets[dsname]
                    else False
                )
                combined_dataset = (
                    datasets[dsname]["combined_dataset"]
                    if "combined_dataset" in datasets[dsname]
                    else None
                )
                # error checking in Dataset

            self.datasets[dsname] = Dataset(
                data=datasets[dsname]["data"],
                model=datasets[dsname]["model"],
                model_parameters=datasets[dsname]["model_parameters"],
                parameters=parameters,
                costfunction=datasets[dsname]["costfunction"],
                name=dsname,
                try_combine=try_combine,
                combined_dataset=combined_dataset,
                user_gradient=user_gradient,
                use_log=use_log,
            )

            for par in datasets[dsname]["model_parameters"].values():
                alldspars.add(par)

        # here is where we should try to combine datasets (or maybe just above?)
        # I *think* we want to keep the datasets as is and maybe add a new holder for
        # the combined datasets and add only the relevant costfunctions together. going to have to look
        # at how constraints are added and covariance matrix made.

        if try_to_combine_datasets:
            # maybe there's more than one combined_dataset group
            for cdsname in combined_datasets:
                # find the Datasets to try to combine
                ds_tocombine = []
                for dsname in self.datasets.keys():
                    if (
                        self.datasets[dsname].try_combine
                        and self.datasets[dsname].combined_dataset == cdsname
                    ):
                        ds_tocombine.append(self.datasets[dsname])

                combined_dataset, included_datasets = combine_datasets(
                    datasets=ds_tocombine,
                    model=combined_datasets[cdsname]["model"],
                    model_parameters=combined_datasets[cdsname]["model_parameters"],
                    parameters=self.parameters,
                    costfunction=combined_datasets[cdsname]["costfunction"],
                    name=cdsname,
                    user_gradient=user_gradient,
                    use_log=use_log,
                )

                if len(included_datasets) > 0:
                    self.combined_datasets[cdsname] = combined_dataset
                    self.included_in_combined_datasets[cdsname] = included_datasets

                for par in combined_datasets[cdsname]["model_parameters"].values():
                    alldspars.add(par)

        # add the costfunctions together
        first = True
        for dsname in self.datasets.keys():
            if self.datasets[dsname].is_combined:
                continue
            if first:
                self.costfunction = self.datasets[dsname].costfunction
                first = False
            else:
                self.costfunction += self.datasets[dsname].costfunction

        for cdsname in self.combined_datasets.keys():
            if first:
                self.costfunction = self.combined_datasets[cdsname].costfunction
                first = False
            else:
                self.costfunction += self.combined_datasets[cdsname].costfunction

        # fitparameters of Superset are a little different than fitparameters of Dataset
        self.fitparameters = self.costfunction._parameters

        # check that parameters are actually used in the Datasets or combined_datasets and remove them if not
        for parameter in list(self.parameters.keys()):
            if parameter not in alldspars:
                msg = f"'{parameter}' included as a parameter but not used in a `Dataset` - removing '{parameter}' as a parameter"
                logging.warning(msg)
                del self.parameters[parameter]

        # collect which parameters are included as parameters to vary for the toys
        for dsname, ds in self.datasets.items():
            for parname in ds._toy_pars_to_vary:
                if parname not in self._toy_pars_to_vary:
                    self._toy_pars_to_vary.append(parname)
                    msg = f"`Superset` '{self.name}' added parameter '{parname}' as a parameter to vary for toys"
                    logging.info(msg)

        if len(self._toy_pars_to_vary) > 0 and constraints is None:
            msg = "have parameters to vary but no constraints found!"
            logging.error(msg)
            raise ValueError(msg)

        if constraints is None:
            msg = "no constraints were provided"
            logging.info(msg)

        if constraints is not None:
            msg = "all constraints will be combined into a single `NormalConstraint`"
            logging.info(msg)

            # shove all the constraints in one big matrix
            for constraintname, constraint in constraints.items():
                # would love to move this somewhere else, maybe sanitize the config before doing anything
                if len(constraint["parameters"]) != len(constraint["values"]):
                    if len(constraint["values"]) == 1:
                        constraint["values"] = np.full(
                            len(constraint["parameters"]), constraint["values"]
                        )
                        msg = f"in constraint '{constraintname}', assigning 1 provided value to all {len(constraint['parameters'])} 'parameters'"
                        logging.warning(msg)
                    else:
                        msg = f"constraint '{constraintname}' has {len(constraint['parameters'])} 'parameters' but {len(constraint['values'])} 'values'"
                        logging.error(msg)
                        raise ValueError(msg)

                if "covariance" in constraint and "uncertainty" in constraint:
                    msg = f"constraint '{constraintname}' has both 'covariance' and 'uncertainty'; this is ambiguous - use only one!"
                    logging.error(msg)
                    raise KeyError(msg)

                if "covariance" not in constraint and "uncertainty" not in constraint:
                    msg = f"constraint '{constraintname}' has neither 'covariance' nor 'uncertainty' - one (and only one) must be provided!"
                    logging.error(msg)
                    raise KeyError(msg)

                # do some cleaning up of the config here
                if "uncertainty" in constraint:
                    if len(constraint["uncertainty"]) > 1:
                        constraint["uncertainty"] = np.full(
                            len(constraint["parameters"]), constraint["uncertainty"]
                        )
                        msg = f"constraint '{constraintname}' has {len(constraint['parameters'])} parameters but only 1 uncertainty - assuming this is constant uncertainty for each parameter"
                        logging.warning(msg)

                    if len(constraint["uncertainty"]) != len(constraint["parameters"]):
                        msg = f"constraint '{constraintname}' has {len(constraint['parameters'])} 'parameters' but {len(constraint['uncertainty'])} 'uncertainty' - should be same length or single uncertainty"
                        logging.error(msg)
                        raise ValueError(msg)

                    # convert to covariance matrix so that we're always working with the same type of object
                    constraint["covariance"] = np.diag(constraint["uncertainty"]) ** 2
                    del constraint["uncertainty"]

                    msg = f"constraint '{constraintname}': converting provided 'uncertainty' to 'covariance'"
                    logging.info(msg)

                else:  # we have the covariance matrix for this constraint
                    if len(constraint["parameters"]) == 1:
                        msg = f"constraint '{constraintname}' has one parameter but uses 'covariance' - taking this at face value"
                        logging.info(msg)

                    if np.shape(constraint["covariance"]) != (
                        len(constraint["parameters"]),
                        len(constraint["parameters"]),
                    ):
                        msg = f"constraint '{constraintname}' has 'covariance' of shape {np.shape(constraint['covariance'])} but it should be shape {(len(constraint['parameters']), len(constraint['parameters']))}"
                        logging.error(msg)
                        raise ValueError(msg)

                    if not np.allclose(
                        constraint["covariance"], np.asarray(constraint["covariance"]).T
                    ):
                        msg = f"constraint '{constraintname}' has non-symmetric 'covariance' matrix - this is not allowed."
                        logging.error(msg)
                        raise ValueError(msg)

                    sigmas = np.sqrt(np.diag(np.asarray(constraint["covariance"])))
                    cov = np.outer(sigmas, sigmas)
                    corr = constraint["covariance"] / cov
                    if not np.all(np.logical_or(np.abs(corr) < 1, np.isclose(corr, 1))):
                        msg = f"constraint '{constraintname}' 'covariance' matrix does not seem to contain proper correlation matrix"
                        logging.error(msg)
                        raise ValueError(msg)

                for par in constraint["parameters"]:
                    if par in self.constraints_parameters:
                        msg = f"parameter {par} is used in multiple constraints - not allowed"
                        logging.error(msg)
                        raise KeyError(msg)

                    self.constraints_parameters[par] = len(self.constraints_parameters)

            # initialize now that we know how large to make them
            self.constraints_values = np.full(len(self.constraints_parameters), np.nan)
            self.constraints_covariance = np.identity(len(self.constraints_parameters))

            for constraintname, constraint in constraints.items():
                # now put the values in
                for par, value in zip(constraint["parameters"], constraint["values"]):
                    self.constraints_values[self.constraints_parameters[par]] = value

                # now put the covariance matrix in
                for i in range(len(constraint["parameters"])):
                    for j in range(len(constraint["parameters"])):
                        self.constraints_covariance[
                            self.constraints_parameters[constraint["parameters"][i]],
                            self.constraints_parameters[constraint["parameters"][j]],
                        ] = constraint["covariance"][i, j]

            self._add_constraints_to_costfunction()  # stored in self.constraints

        # so we can mutate self._toy_parameters later without it being a problem?
        self._toy_parameters = deepcopy(self.parameters)

        # that's all, folks!

    def _add_constraints_to_costfunction(
        self,
    ) -> None:
        if self.constraints is not None:
            msg = "already added constraints to costfunction! not adding again"
            logging.warning(msg)
            return self.constraints

        pars, values, covariance = self.get_constraints(self.fitparameters)

        self.constraints = cost.NormalConstraint(pars, values, error=covariance)

        self.costfunction = self.costfunction + self.constraints

        return

    # gets the appropriate values and covariance submatrix for the requested parameters, if they exist
    # returns a tuple that contains a list of parameters found constraints for, their values, covariance matrix
    def get_constraints(
        self,
        parameters: list,
    ) -> tuple:
        if len(self.constraints_parameters) == 0:
            return (None, None, None)

        pars = []
        inds = []
        for par in parameters:
            if par in self.constraints_parameters:
                pars.append(par)
                inds.append(self.constraints_parameters[par])

        values = self.constraints_values[inds]
        covar = self.constraints_covariance[np.ix_(inds, inds)]

        return (pars, values, covar)



import numba as nb
import numpy as np

import freqfit.models.constants as constants
from freqfit.utils import inspectparameters

nb_kwd = {
    "nopython": True,
    "parallel": False,
    "nogil": True,
    "cache": True,
    "fastmath": True,
    "inline": "always",
}

QBB = constants.QBB
N_A = constants.NA
M_A = constants.MDET

# default analysis window and width
# window
#     uniform background regions to pull from, must be a 2D array of form e.g. `np.array([[0,1],[2,3]])`
#     where edges of window are monotonically increasing (this is not checked), in keV.
#     Default is typical analysis window.
WINDOW = np.array(constants.WINDOW)

WINDOWSIZE = 0.0
for i in range(len(WINDOW)):
    WINDOWSIZE += WINDOW[i][1] - WINDOW[i][0]

FULLWINDOWSIZE = WINDOW[-1][1] - WINDOW[0][0]


SEED = 42  # set the default random seed


@nb.jit(**nb_kwd)
def nb_pdf(
    Es: np.array,
    S: float,
    BI: float,
    delta: float,
    sigma: float,
    eff: float,
    effunc: float,
    effuncscale: float,
    exp: float,
    a: float,
    check_window: bool = False,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    S
        The signal rate, in units of counts/(kg*yr)
    BI
        The background index rate, in counts/(kg*yr*keV)
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    eff
        The global signal efficiency, unitless
    effunc
        uncertainty on the efficiency
    effuncscale
        scaling parameter of the efficiency
    exp
        The exposure, in kg*yr
    a
        Controls the decay constant of the exponential background
    check_window
        whether to check if the passed Es fall inside of the window. Default is False and assumes that the passed Es
        all fall inside the window (for speed)

    Notes
    -----
    This function computes the following:
    mu_S = (eff + effuncscale * effunc) * exp * S
    mu_B = BI * W * exp
    pdf(E) = 1/(mu_S+mu_B) * [mu_S * norm(E_j, QBB - delta, sigma) + N*BI*W*exp*(np.exp(-a*(E_j-x_lo)/a + 1))]
    """
    x_lo = WINDOW[0][0]
    x1 = WINDOW[0][1]
    x2 = WINDOW[1][0]
    x3 = WINDOW[1][1]
    x4 = WINDOW[-1][0]
    x_hi = WINDOW[-1][1]

    # Compute the normalization for the CDF
    invnorm = (x_hi - x4 + x3 - x2 + x1 - x_lo) - (
        np.exp(-a * (x_hi - x_lo))
        - np.exp(-a * (x4 - x_lo))
        + np.exp(-a * (x3 - x_lo))
        - np.exp(-a * (x2 - x_lo))
        + np.exp(-a * (x1 - x_lo))
        - 1
    )

    if invnorm == 0:
        norm = np.inf
    else:
        norm = 1 / invnorm

    if a == 0:
        a_inv = np.inf
    else:
        a_inv = 1 / a

    # Precompute the signal and background counts
    # mu_S = np.log(2) * (N_A * S) * (eff + effuncscale * effunc) * exp / M_A
    mu_S = S * (eff + effuncscale * effunc) * exp
    mu_B = WINDOWSIZE * BI * exp

    # Precompute the prefactors so that way we save multiplications in the for loop
    S_amp = mu_S / (np.sqrt(2 * np.pi) * sigma)
    B_amp = norm * mu_B

    # Initialize and execute the for loop
    y = np.empty_like(Es, dtype=np.float64)
    for i in nb.prange(Es.shape[0]):
        y[i] = (1 / (mu_S + mu_B)) * (
            S_amp * np.exp(-((Es[i] - QBB + delta) ** 2) / (2 * sigma**2))
            + B_amp * (np.exp(-a * (Es[i] - x_lo)) * a_inv + 1)
        )

    if check_window:
        for i in nb.prange(Es.shape[0]):
            inwindow = False
            for j in range(len(WINDOW)):
                if WINDOW[j][0] <= Es[i] <= WINDOW[j][1]:
                    inwindow = True
            if not inwindow:
                y[i] = 0.0

    return y


@nb.jit(**nb_kwd)
def nb_density(
    Es: np.array,
    S: float,
    BI: float,
    delta: float,
    sigma: float,
    eff: float,
    effunc: float,
    effuncscale: float,
    exp: float,
    a: float,
    check_window: bool = True,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    S
        The signal rate, in units of counts/(kg*yr)
    BI
        The background index rate, in counts/(kg*yr*keV)
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    eff
        The global signal efficiency, unitless
    effunc
        uncertainty on the efficiency
    effuncscale
        scaling parameter of the efficiency
    exp
        The exposure, in kg*yr
    a
        Controls the decay constant of the exponential background

    Notes
    -----
    This function computes the following, faster than without a numba wrapper:
    mu_S = (eff + effuncscale * effunc) * exp * S
    mu_B = BI * W * exp
    CDF(E) = mu_S + mu_B
    pdf(E) =[mu_S * norm(E_j, QBB + delta, sigma) + N * BI * W * exp * (np.exp(-a*(E_j-x_lo))/a + 1)]
    """

    x_lo = WINDOW[0][0]
    x1 = WINDOW[0][1]
    x2 = WINDOW[1][0]
    x3 = WINDOW[1][1]
    x4 = WINDOW[-1][0]
    x_hi = WINDOW[-1][1]

    # Compute the normalization for the CDF
    invnorm = (x_hi - x4 + x3 - x2 + x1 - x_lo) - (
        np.exp(-a * (x_hi - x_lo))
        - np.exp(-a * (x4 - x_lo))
        + np.exp(-a * (x3 - x_lo))
        - np.exp(-a * (x2 - x_lo))
        + np.exp(-a * (x1 - x_lo))
        - 1
    )

    if invnorm == 0:
        norm = np.inf
    else:
        norm = 1 / invnorm

    if a == 0:
        a_inv = np.inf
    else:
        a_inv = 1 / a

    mu_S = S * (eff + effuncscale * effunc) * exp
    mu_B = WINDOWSIZE * BI * exp

    if sigma == 0:
        return np.inf, np.full_like(Es, np.inf, dtype=np.float64)

    # Precompute the prefactors so that way we save multiplications in the for loop
    S_amp = mu_S / (np.sqrt(2 * np.pi) * sigma)
    B_amp = norm * mu_B

    # Initialize and execute the for loop
    y = np.empty_like(Es, dtype=np.float64)
    for i in nb.prange(Es.shape[0]):
        y[i] = S_amp * np.exp(
            -((Es[i] - QBB + delta) ** 2) / (2 * sigma**2)
        ) + B_amp * (np.exp(-a * (Es[i] - x_lo)) * a_inv + 1)

    if check_window:
        for i in nb.prange(Es.shape[0]):
            inwindow = False
            for j in range(len(WINDOW)):
                if WINDOW[j][0] <= Es[i] <= WINDOW[j][1]:
                    inwindow = True
            if not inwindow:
                y[i] = 0.0

    return mu_S + mu_B, y


@nb.jit(**nb_kwd)
def nb_density_gradient(
    Es: np.array,
    S: float,
    BI: float,
    delta: float,
    sigma: float,
    eff: float,
    effunc: float,
    effuncscale: float,
    exp: float,
    m: float,
) -> np.array:
    raise NotImplementedError
    return


@nb.jit(**nb_kwd)
def nb_logpdf(
    Es: np.array,
    S: float,
    BI: float,
    delta: float,
    sigma: float,
    eff: float,
    effunc: float,
    effuncscale: float,
    exp: float,
    m: float,
) -> np.array:
    raise NotImplementedError
    return


@nb.jit(nopython=True, fastmath=True, cache=True, error_model="numpy")
def nb_rvs(
    n_sig: int,
    n_bkg: int,
    delta: float,
    sigma: float,
    seed: int = SEED,
) -> np.array:
    raise NotImplementedError
    return


@nb.jit(nopython=True, fastmath=True, cache=True, error_model="numpy")
def nb_extendedrvs(
    S: float,
    BI: float,
    delta: float,
    sigma: float,
    eff: float,
    effunc: float,
    effuncscale: float,
    exp: float,
    a: float,
    seed: int = SEED,
) -> np.array:
    """
    Parameters
    ----------
    S
        expected rate of signal events in events/(kg*yr)
    BI
        rate of background events in events/(kev*kg*yr)
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    eff
        The global signal efficiency, unitless
    effunc
        uncertainty on the efficiency
    effuncscale
        scaling parameter of the efficiency
    exp
        The exposure, in kg*yr
    a
        Controls the decay of the exponential background
    seed
        specify a seed, otherwise uses default seed

    Notes
    -----
    This function pulls from a Gaussian for signal events and from a uniform distribution for background events
    in the provided windows, which may be discontinuous.
    """
    raise NotImplementedError
    # S *= 0.01
    # BI *= 0.0001

    # np.random.seed(seed)

    # n_sig = np.random.poisson(S * (eff + effuncscale * effunc) * exp)
    # n_bkg = np.random.poisson(BI * exp * WINDOWSIZE)

    # x_lo = WINDOW[0][0]
    # x1 = WINDOW[0][1]
    # x2 = WINDOW[1][0]
    # x3 = WINDOW[1][1]
    # x4 = WINDOW[-1][0]
    # x_hi = WINDOW[-1][1]

    # # Compute the normalization for the CDF
    # invnorm = (x_hi - x4 + x3 - x2 + x1 -x_lo) - (np.exp(-a*(x_hi-x_lo)) - np.exp(-a*(x4-x_lo)) + np.exp(-a*(x3-x_lo)) - np.exp(-a*(x2-x_lo))+
    # np.exp(-a*(x1-x_lo)) - 1)

    # if invnorm == 0:
    #     norm = np.inf
    # else:
    #     norm = 1/invnorm

    # Es = np.random.exponential(a, n_bkg) + x_lo
    # # Make sure we drew in the correct window
    # for i in nb.prange(len(Es)):
    #     inwindow = False
    #     for j in range(len(WINDOW)):
    #         if WINDOW[j][0] <= Es[i] <= WINDOW[j][1]:
    #             inwindow = True
    #     if inwindow:
    #         # loop until we do get a count inside a window
    #         new_inwindow = True
    #         while new_inwindow:
    #             newdraw = np.random.exponential(a, 1)[0] + x_lo
    #             new_inwindowcheck = False
    #             for j in range(len(WINDOW)-1):
    #                 if WINDOW[j][1] <= newdraw <= WINDOW[j+1][0]:
    #                     new_inwindowcheck = True
    #             if not new_inwindowcheck:
    #                 new_inwindow = False
    #     Es[i] = newdraw

    #     Es = np.append(Es, np.random.normal(QBB - delta, sigma, size=n_sig))

    # return Es, (n_bkg, n_sig)


class correlated_efficiency_0vbb_exponential_background_gen:
    def __init__(self):
        self.parameters = inspectparameters(self.density)
        pass

    def pdf(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        effunc: float,
        effuncscale: float,
        exp: float,
        a: float,
        check_window: bool = False,
    ) -> np.array:
        return nb_pdf(
            Es, S, BI, delta, sigma, eff, effunc, effuncscale, exp, a, check_window
        )

    def logpdf(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        effunc: float,
        effuncscale: float,
        exp: float,
        a: float,
    ) -> np.array:
        return nb_logpdf(Es, S, BI, delta, sigma, eff, effunc, effuncscale, exp, a)

    # for iminuit ExtendedUnbinnedNLL
    def density(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        effunc: float,
        effuncscale: float,
        exp: float,
        a: float,
    ) -> np.array:
        return nb_density(Es, S, BI, delta, sigma, eff, effunc, effuncscale, exp, a)

    # for iminuit ExtendedUnbinnedNLL
    def density_gradient(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        effunc: float,
        effuncscale: float,
        exp: float,
        a: float,
    ) -> np.array:
        return nb_density_gradient(
            Es, S, BI, delta, sigma, eff, effunc, effuncscale, exp, a
        )

    # for iminuit ExtendedUnbinnedNLL
    def log_density(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        effunc: float,
        effuncscale: float,
        exp: float,
        a: float,
    ) -> np.array:
        mu_S = S * (eff + effuncscale * effunc) * exp
        mu_B = exp * BI * WINDOWSIZE

        # Do a quick check and return -inf if log args are negative
        if (mu_S + mu_B <= 0) or np.isnan(np.array([mu_S, mu_B])).any():
            return mu_S + mu_B, np.full(Es.shape[0], -np.inf)
        else:
            return (
                mu_S + mu_B,
                np.log(mu_S + mu_B)
                + nb_logpdf(Es, S, BI, delta, sigma, eff, effunc, effuncscale, exp, a),
            )

    # should we have an rvs method for drawing a random number of events?
    # `extendedrvs`
    # needs to use same parameters as the rest of the functions...
    def rvs(
        self,
        n_sig: int,
        n_bkg: int,
        delta: float,
        sigma: float,
        a: float,
        seed: int = SEED,
    ) -> np.array:
        return nb_rvs(n_sig, n_bkg, delta, sigma, a, seed=seed)

    def extendedrvs(
        self,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        effunc: float,
        effuncscale: float,
        exp: float,
        a: float,
        seed: int = SEED,
    ) -> np.array:
        return nb_extendedrvs(
            S, BI, delta, sigma, eff, effunc, effuncscale, exp, a, seed=seed
        )

    def plot(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        effunc: float,
        effuncscale: float,
        exp: float,
        a: float,
    ) -> None:
        y = nb_pdf(Es, S, BI, delta, sigma, eff, effunc, effuncscale, exp, a)

        import matplotlib.pyplot as plt

        plt.step(Es, y)
        plt.show()

    # function call needs to take the same parameters as the other function calls, in the same order repeated twice
    # this is intended only for empty datasets
    # returns `None` if we couldn't combine the datasets (a dataset was not empty)
    def combine(
        self,
        a_Es: np.array,
        a_S: float,
        a_BI: float,
        a_delta: float,
        a_sigma: float,
        a_eff: float,
        a_effunc: float,
        a_effuncscale: float,
        a_exp: float,
        a_m: float,
        b_Es: np.array,
        b_S: float,
        b_BI: float,
        b_delta: float,
        b_sigma: float,
        b_eff: float,
        b_effunc: float,
        b_effuncscale: float,
        b_exp: float,
        b_m: float,
    ) -> list | None:
        # datasets must be empty to be combined
        if len(a_Es) != 0 or len(b_Es) != 0:
            return None

        Es = np.array([])  # both of these datasets are empty
        S = 0.0  # this should be overwritten in the fit later
        BI = 0.0  # this should be overwritten in the fit later
        m = 0.0  # this should be overwritten in the fit later

        exp = a_exp + b_exp  # total exposure

        # exposure weighted fixed parameters (important to calculate correctly)
        sigma = (a_exp * a_sigma + b_exp * b_sigma) / exp
        eff = (a_exp * a_eff + b_exp * b_eff) / exp
        delta = (a_exp * a_delta + b_exp * b_delta) / exp

        # these are fully correlated in this model so the direct sum is appropriate
        # (maybe still appropriate even if not fully correlated?)
        effunc = (a_exp * a_effunc + b_exp * b_effunc) / exp

        effuncscale = 0.0  # this should be overwritten in the fit later

        return [Es, S, BI, delta, sigma, eff, effunc, effuncscale, exp, m]

    def can_combine(
        self,
        a_Es: np.array,
        a_S: float,
        a_BI: float,
        a_delta: float,
        a_sigma: float,
        a_eff: float,
        a_effunc: float,
        a_effuncscale: float,
        a_exp: float,
        a_m: float,
    ) -> bool:
        """
        This sets an arbitrary rule if this dataset can be combined with other datasets.
        In this case, if the dataset contains no data, then it can be combined, but more complex rules can be imposed.
        """
        if len(a_Es) == 0:
            return True
        else:
            return False


correlated_efficiency_0vbb_exponential_background = (
    correlated_efficiency_0vbb_exponential_background_gen()
)



"""
Plotting utilities for outputs of `SetLimit`
"""
import glob
import logging
from math import erf

import h5py
import legendstyles
import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import chi2

from legendfreqfit.models.constants import (
    NME_PHENO_HIGH,
    NME_PHENO_LOW,
    NME_central,
    NME_unc_hi,
    NME_unc_lo,
    m_prime_to_m,
    s_prime_to_s,
)
from legendfreqfit.statistics import (
    find_crossing,
    get_p_values,
    sensitivity,
    ts_critical,
)

plt.style.use(legendstyles.LEGEND)

import matplotlib.pyplot as plt
from matplotlib.collections import PatchCollection
from matplotlib.ticker import MultipleLocator


# define an object that will be used by the legend
class MulticolorPatch:
    def __init__(self, colors):
        self.colors = colors


# define a handler for the MulticolorPatch object
class MulticolorPatchHandler:
    def legend_artist(self, legend, orig_handle, fontsize, handlebox):
        width, height = handlebox.width, handlebox.height
        patches = []
        for i, c in enumerate(orig_handle.colors):
            patches.append(
                plt.Rectangle(
                    [
                        width / len(orig_handle.colors) * i - handlebox.xdescent,
                        -handlebox.ydescent,
                    ],
                    width / len(orig_handle.colors),
                    height,
                    facecolor=c,
                    edgecolor="none",
                )
            )

        patch = PatchCollection(patches, match_original=True)

        handlebox.add_artist(patch)
        return patch


PLTSIZE = legendstyles.figsizes.JupyterNotebook

GOOD_ORANGE = "#ff5d07"
GOOD_MAGENTA = "#D907FF"
NICE_BLUE = "#668DA5"
NICE_RED = "#B4584D"
NICE_GREEN = "#ABB1A2"
NICE_PINK = "#CCACAD"


BRAZIL_1_SIGMA_KWARGS = {"color": "#07A9FF", "alpha": 1}
BRAZIL_2_SIGMA_KWARGS = {"color": "#07A9FF", "alpha": 0.4}
P_VALUE_KWARGS = {"color": "grey", "alpha": 1}

log = logging.getLogger(__name__)


class PlotLimit:
    def __init__(
        self,
        ts_dir: str,
        brazil_dir: str,
        TRUNC=False,
        TRUNC_SIZE=11,
        RANDOM=False,
        NME=False,
        RV_SIZE=4,
    ):
        """
        ts_dir
            Absolute path to the directory containing the test statistic files generated with and tested against signal
        brazil_dir
            Absolute path to the directory containing the test statistic files generated with no signal and tested against signal
        """
        self.RANDOM = RANDOM
        self.TRUNC = TRUNC
        self.TRUNC_SIZE = TRUNC_SIZE
        self.NME = NME
        f = glob.glob(ts_dir + "/*_wilks_fine.h5")[0]

        if self.NME:
            df = h5py.File(f, "r")
            s = df["m_bb"][:]
            s_scanned = s
            ts = df["t_m_bb"][:]
            if ts.shape[1] > 1:
                ts = ts[:, 0]
            df.close()

            self.signal_rate = m_prime_to_m(s)
            self.ts = ts

        else:
            df = h5py.File(f, "r")
            s = df["s"][:]
            s_scanned = s
            ts = df["t_s"][:]
            if ts.shape[1] > 1:
                ts = ts[:, 0]
            df.close()

            self.signal_rate = s_prime_to_s(s)
            self.ts = ts

        files_per_s = []

        for s_val in s_scanned[:]:
            files = glob.glob(ts_dir + f"/{s_val}_*.h5")
            files_per_s.append(files)

        t_crits = []
        t_crit_lows = []
        t_crit_his = []

        toys_per_scanned_s = []
        Es_per_scanned_s = []

        for file_list in files_per_s[:]:
            toy_ts_per_file = []
            toy_num_per_file = []
            toy_denom_per_file = []
            Es_per_file = []
            try:
                if self.RANDOM:
                    for FILE in file_list:
                        f = h5py.File(FILE, "r")
                        if "Es" in f.keys():
                            Es_per_file.extend(f["Es"][:])
                        else:
                            Es_per_file.extend([])
                        toy_ts_per_file.extend(
                            f["ts"][:] + np.random.normal(0, RV_SIZE, len(f["ts"][:]))
                        )
                        f.close()
                elif self.TRUNC:
                    for FILE in file_list:
                        f = h5py.File(FILE, "r")
                        if "Es" in f.keys():
                            Es_per_file.extend(f["Es"][:])
                        else:
                            Es_per_file.extend([])
                        toy_ts_per_file.extend(np.around(f["ts"][:], self.TRUNC_SIZE))
                        toy_num_per_file.extend(
                            np.around(f["ts_num"][:], self.TRUNC_SIZE)
                        )
                        toy_denom_per_file.extend(
                            np.around(f["ts_denom"][:], self.TRUNC_SIZE)
                        )
                        f.close()
                else:
                    for FILE in file_list:
                        f = h5py.File(FILE, "r")
                        toy_ts_per_file.extend(f["ts"][:])
                        if "Es" in f.keys():
                            Es = f["Es"][:]
                            num = f["ts_num"][:]
                            denom = f["ts_denom"][:]
                        else:
                            Es = []
                            num = []
                            denom = []
                        Es_per_file.extend(Es)
                        toy_num_per_file.extend(num)
                        toy_denom_per_file.extend(denom)
                        f.close()

                tcrit_tuple, _ = ts_critical(
                    toy_ts_per_file, threshold=0.9, confidence=0.68, plot=False
                )
                t_crit, t_crit_low, t_crit_high = tcrit_tuple

                t_crits.append(t_crit)
                t_crit_lows.append(t_crit_low)
                t_crit_his.append(t_crit_high)

                toys_per_scanned_s.append(toy_ts_per_file)

                Es_per_scanned_s.append(Es_per_file)
            except:  # noqa: B001, E722
                t_crits.append(np.nan)
                t_crit_lows.append(np.nan)
                t_crit_his.append(np.nan)

        self.toys_per_scanned_s = toys_per_scanned_s
        self.Es_per_scanned_s = Es_per_scanned_s
        self.t_crits = t_crits
        self.t_crit_lows = t_crit_lows
        self.t_crit_his = t_crit_his

        self.p_values = get_p_values(toys_per_scanned_s, ts)

        brazil_files = glob.glob(brazil_dir + "/*.h5")

        toys_per_s_zero_sig = []
        num_per_s_zero_sig = []
        denom_per_s_zero_sig = []
        Es_per_s_zero_sig = []
        seeds_per_s_zero_sig = []
        for i in range(len(s[:])):
            toys_per_file = []
            num_per_file = []
            denom_per_file = []
            Es_per_file = []
            seeds_per_file = []

            if self.RANDOM:
                for file in brazil_files:
                    f = h5py.File(file, "r")
                    ts = f["ts"][:]
                    Es = f["Es"][:]
                    seeds = f["seed"][:]
                    toys_per_file.extend(
                        ts[i] + np.random.normal(0, RV_SIZE, len(ts[i]))
                    )
                    Es_per_file.extend(Es)
                    seeds_per_file.extend(seeds)

            elif self.TRUNC:
                for file in brazil_files:
                    f = h5py.File(file, "r")
                    ts = f["ts"][:]
                    Es = f["Es"][:]
                    seeds = f["seed"][:]
                    toys_per_file.extend(np.around(ts[i], TRUNC_SIZE))
                    denom_per_file.extend(np.around(f["ts_denom"][:][i], TRUNC_SIZE))
                    num_per_file.extend(np.around(f["ts_num"][:][i], TRUNC_SIZE))
                    Es_per_file.extend(Es)
                    seeds_per_file.extend(seeds)
            else:
                for file in brazil_files:
                    f = h5py.File(file, "r")
                    ts = f["ts"][:]
                    toys_per_file.extend(ts[i])
                    if "Es" in f.keys():
                        Es = f["Es"][:]
                        seeds = f["seed"][:]
                        num = f["ts_num"][:][i]
                        denom = f["ts_denom"][:][i]
                    else:
                        Es = []
                        seeds = []
                        denom = []
                        num = []

                    Es_per_file.extend(Es)
                    seeds_per_file.extend(seeds)
                    denom_per_file.extend(denom)
                    num_per_file.extend(num)
            toys_per_s_zero_sig.append(toys_per_file)
            num_per_s_zero_sig.append(num_per_file)
            denom_per_s_zero_sig.append(denom_per_file)
            Es_per_s_zero_sig.append(Es_per_file)
            seeds_per_s_zero_sig.append(seeds_per_file)

        self.toys_per_s_zero_sig = toys_per_s_zero_sig
        self.num_per_s_zero_sig = num_per_s_zero_sig
        self.denom_per_s_zero_sig = denom_per_s_zero_sig
        self.Es_per_s_zero_sig = Es_per_s_zero_sig
        self.seeds_per_s_zero_sig = seeds_per_s_zero_sig

        self.p_values_median, self.p_values_hi, self.p_values_lo = sensitivity(
            toys_per_s_zero_sig[:],
            toys_per_scanned_s[:],
            s_scanned[:],
            CL=erf(2 / np.sqrt(2)),
            plot=False,
        )
        self.p_values_median, self.p_values_hi_1, self.p_values_lo_1 = sensitivity(
            toys_per_s_zero_sig[:],
            toys_per_scanned_s[:],
            s_scanned[:],
            CL=erf(1 / np.sqrt(2)),
            plot=False,
        )

        # Some plotting defaults that the user can override before function calls
        # Defaults for critical ts plot
        self.critical_y_lim = (-0.25, 6.25)
        self.critical_x_lim = (-1e-2, self.signal_rate[-1] / 1e-25)

        # Defaults for sensitivity plot
        self.sensitivity_y_lim = (1e-3, 1.2)
        self.sensitivity_x_lim = (-1e-3, self.signal_rate[-1] / 1e-25)

        # Defaults for test statistic plot
        self.idx = 19
        self.RANGE = (-1, 10)

        # For NME sensitivity
        self.plot_nme_range = False
        self.nme_sensitivity_x_lim = (-1e-3, self.signal_rate[-1] * 1000)

    def plot_critical_test_statistic(self):
        if self.NME:
            s_approx = find_crossing(self.signal_rate, self.t_crits, self.ts)
            T_est = s_approx[-1]
            gammas = self.signal_rate
        else:
            gammas = self.signal_rate / 1e-25
            s_approx = find_crossing(gammas, self.t_crits, self.ts)
            T_est = 1 / (s_approx[-1] * 1e-25)

        fig, ax = plt.subplots(figsize=legendstyles.figsizes.JupyterNotebook)
        plt.axhline(y=2.71, label="Wilks' 90% CL", ls=":", **P_VALUE_KWARGS)

        plt.axvline(x=s_approx[0], label="90% CL crossing", ls=":", c="k")
        plt.axvline(x=s_approx[-1], ls=":", c="k")

        plt.plot(
            gammas,
            self.ts,
            ls="None",
            marker=".",
            color="k",
            label=r"$\tilde{t}_{\Gamma, \mathrm{ obs}}$",
        )
        plt.plot(
            gammas,
            self.t_crits,
            c=NICE_RED,
            label=r"$\tilde{t}_{\Gamma, 90\%\, \mathrm{ CL}}$",
        )
        plt.fill_between(
            gammas, self.t_crit_lows, self.t_crit_his, alpha=0.6, color=NICE_RED
        )

        if self.NME:
            plt.xlabel(r"$m_{\beta\beta}$ [eV]")
            plt.ylabel(r"$\tilde{t}_{m_{\beta\beta}}$")
            plt.plot(
                [],
                [],
                ls="None",
                label=r"limit $m_{\beta\beta} <$" + f"{s_approx[-1]: .2f} eV",
            )
        else:
            plt.xlabel(r"$\Gamma_{1/2}^{0\nu}$ [$10^{-25} \,\mathrm{yr}^{-1}$]")
            plt.ylabel(r"$\tilde{t}_{\Gamma}$")
            plt.plot(
                [],
                [],
                ls="None",
                label=rf"$\rm{{T}}_{{1/2}}^{{0\nu}} >${T_est: 0.2e} yr, 90% CL",
            )

        legendstyles.legend_watermark(ax, logo_suffix="-200", approved=True)
        # legendstyles.add_preliminary(ax, color="red")

        plt.ylim(*self.critical_y_lim)
        plt.xlim(*self.critical_x_lim)

        plt.legend()
        plt.tight_layout()

        return fig

    def plot_sensitivity(self):
        s_approx = find_crossing(self.signal_rate, self.p_values, 0.1)
        T_est = 1 / s_approx[-1]
        s_approx = find_crossing(self.signal_rate, self.p_values_median, 0.1)
        T_12_median = 1 / s_approx[-1]
        gammas = self.signal_rate / 1e-25

        fig, ax = plt.subplots(figsize=legendstyles.figsizes.JupyterNotebook)
        plt.plot(
            gammas,
            self.p_values,
            marker="s",
            markersize=3,
            color="k",
            label="obs., " + r"$\rm{T}_{1/2}^{0\nu} >$ " + f"{T_est: 0.2e} yr, 90% CL",
        )
        plt.plot(
            gammas,
            self.p_values_median,
            color="k",
            ls="--",
            label="median exp. bkg. only, "
            + r"$\rm{T}_{1/2}^{0\nu} >$ "
            + f"{T_12_median: 0.2e} yr, 90% CL",
        )

        plt.fill_between(
            gammas,
            self.p_values_lo,
            self.p_values_hi,
            **BRAZIL_2_SIGMA_KWARGS,
        )
        plt.fill_between(
            gammas,
            self.p_values_lo_1,
            self.p_values_hi_1,
            **BRAZIL_1_SIGMA_KWARGS,
        )

        log.warning(find_crossing(self.signal_rate, self.p_values_lo_1, 0.1))
        log.warning(find_crossing(self.signal_rate, self.p_values_hi_1, 0.1))
        # ------ get the legend-entries that are already attached to the axis
        colors, label = ax.get_legend_handles_labels()

        # ------ append the multicolor legend patches
        colors.append(
            MulticolorPatch(
                [
                    (BRAZIL_1_SIGMA_KWARGS["color"], BRAZIL_1_SIGMA_KWARGS["alpha"]),
                    (BRAZIL_2_SIGMA_KWARGS["color"], BRAZIL_2_SIGMA_KWARGS["alpha"]),
                ]
            )
        )
        label.append(r"median exp. bkg. only $\pm$ 1 $\sigma,\,\pm 2 \sigma$")

        plt.axhline(y=0.1, ls=":", **P_VALUE_KWARGS)
        plt.vlines(1 / T_est / 1e-25, 0, 1e-1, **P_VALUE_KWARGS)
        plt.vlines(1 / T_12_median / 1e-25, 0, 1e-1, ls="--", **P_VALUE_KWARGS)

        ax.xaxis.set_minor_locator(MultipleLocator(5))
        ax.minorticks_on()

        plt.yscale("log")
        plt.xlabel(r"$\Gamma_{1/2}^{0\nu} \, [10^{-25} \,\mathrm{yr}^{-1}]$ ")
        plt.ylabel(r"$p$-value")

        # plt.legend()
        ax.legend(
            colors,
            label,
            loc="best",
            handler_map={MulticolorPatch: MulticolorPatchHandler()},
            frameon=False,
        )
        legendstyles.legend_watermark(ax, logo_suffix="-200", approved=True)
        # legendstyles.add_preliminary(ax, color="red")

        plt.ylim(*self.sensitivity_y_lim)
        plt.xlim(*self.sensitivity_x_lim)
        plt.tight_layout()
        return fig

    def plot_test_statistic(self):
        RANGE = self.RANGE

        plt.figure(figsize=(12, 8))
        n, bins, _ = plt.hist(
            self.toys_per_scanned_s[self.idx], range=RANGE, bins=1000, color="orange"
        )
        n2, bins2, _ = plt.hist(
            self.toys_per_s_zero_sig[self.idx], range=RANGE, bins=1000, color="blue"
        )
        plt.clf()

        fig, ax = plt.subplots(figsize=PLTSIZE)
        gammas = self.signal_rate / 1e-25

        dummy_ts = np.linspace(RANGE[0], RANGE[-1], 100)
        plt.plot(
            dummy_ts,
            chi2.pdf(dummy_ts, df=1)
            * len(self.toys_per_scanned_s[self.idx])
            * (bins[1] - bins[0]),
            ls=":",
            color="grey",
            zorder=2,
            label=r"$\chi^2$ ndof=1",
        )

        plt.axvline(
            np.median(self.toys_per_s_zero_sig[self.idx]),
            label=r"median $\Gamma_0$",
            color="k",
            ls="--",
            zorder=2,
        )
        CL = erf(1 / np.sqrt(2))
        upper_ts = np.quantile(
            self.toys_per_s_zero_sig[self.idx], 0.5 + CL / 2, method="linear"
        )
        lower_ts = np.quantile(
            self.toys_per_s_zero_sig[self.idx], 0.5 - CL / 2, method="linear"
        )
        plt.axvline(upper_ts, label=r"1$\sigma$", ls=":", color="k", alpha=1, zorder=2)
        plt.axvline(lower_ts, ls=":", color="k", alpha=1, zorder=2)

        plt.stairs(
            n2,
            bins2,
            color=legendstyles.colors.AchatBlue,
            label=rf"$\Gamma$'={gammas[self.idx]*1e-25:.4e}|$\Gamma$=0",
            alpha=1,
            zorder=100,
        )
        plt.stairs(
            n,
            bins,
            color=GOOD_ORANGE,
            label=rf"$\Gamma$'={gammas[self.idx]*1e-25:.4e}|$\Gamma$={gammas[self.idx]*1e-25:.4e}",
            alpha=1,
            zorder=100,
        )
        plt.xlabel(r"$\tilde{t}_{\Gamma}$")
        plt.ylabel("counts")
        plt.yscale("log")

        legendstyles.legend_watermark(ax, logo_suffix="-200", approved=True)
        # legendstyles.add_preliminary(ax, color="red")

        plt.legend()
        plt.xlim(*RANGE)

        median_ts = np.median(self.toys_per_s_zero_sig[self.idx])
        log.debug(median_ts)
        ordered_ts = np.sort(self.toys_per_scanned_s[self.idx])
        log.debug((len(ordered_ts[ordered_ts >= median_ts])) / len(ordered_ts))
        plt.tight_layout()

        return fig

    def plot_NME_sensitivity(self):  # noqa: N802
        s_approx = find_crossing(self.signal_rate, self.p_values, 0.1)
        T_est = s_approx[-1]
        s_approx = find_crossing(self.signal_rate, self.p_values_median, 0.1)
        T_12_median = s_approx[-1]
        m_bb = self.signal_rate * 1000

        fig, ax = plt.subplots(figsize=legendstyles.figsizes.JupyterNotebook)
        if self.plot_nme_range:
            (obs,) = ax.plot(
                m_bb,
                self.p_values,
                color="k",
            )
        else:
            (obs,) = ax.plot(m_bb, self.p_values, color="k", marker="s", markersize=3)

        plt.plot(
            m_bb,
            self.p_values_median,
            color="k",
            alpha=1,
            ls="--",
            label="median exp. bkg. only, "
            + r"$m_{\beta\beta} <$ "
            + f"{round(T_12_median*1000,-1): 0.0f} meV, 90% CL",
        )

        plt.fill_between(
            m_bb,
            self.p_values_lo,
            self.p_values_hi,
            **BRAZIL_2_SIGMA_KWARGS,
        )
        plt.fill_between(
            m_bb,
            self.p_values_lo_1,
            self.p_values_hi_1,
            **BRAZIL_1_SIGMA_KWARGS,
        )
        log.warning(find_crossing(self.signal_rate, self.p_values_lo_1, 0.1))
        log.warning(find_crossing(self.signal_rate, self.p_values_hi_1, 0.1))

        if self.plot_nme_range:
            m_bb_belley_low = T_est * 1000 * NME_central / (NME_central + NME_unc_hi)
            m_bb_belley_high = T_est * 1000 * NME_central / (NME_central - NME_unc_lo)
            m_bb_pheno_low = (NME_central / (NME_PHENO_LOW)) * T_est * 1000
            m_bb_pheno_high = (NME_central / (NME_PHENO_HIGH)) * T_est * 1000

            plt.plot(
                m_bb * NME_central / (NME_central + NME_unc_hi),
                self.p_values,
                color=GOOD_MAGENTA,
                alpha=0.5,
            )
            plt.plot(
                m_bb * NME_central / (NME_central - NME_unc_lo),
                self.p_values,
                color=GOOD_MAGENTA,
                alpha=0.5,
            )
            # plot the Belley
            band = ax.fill_between(
                np.linspace(m_bb_belley_low, m_bb_belley_high, 100),
                0.08,
                0.12,
                color=GOOD_MAGENTA,
                hatch="\\\\",
                alpha=0.8,
                facecolor="None",
                linewidth=2,
            )

            plt.vlines(m_bb_belley_low, 0, 1e-1, color=GOOD_MAGENTA, alpha=0.8)
            plt.vlines(m_bb_belley_high, 0, 1e-1, color=GOOD_MAGENTA, alpha=0.8)

            # plot the pheno
            plt.fill_between(
                np.linspace(m_bb_pheno_low, m_bb_pheno_high, 100),
                0.08,
                0.12,
                color=GOOD_ORANGE,
                hatch="//",
                alpha=0.8,
                facecolor="None",
                linewidth=2,
                label="phenom., "
                + r"$m_{\beta\beta} <$ "
                + rf"${round(m_bb_pheno_high, -1): 0.0f}-{round(m_bb_pheno_low, -1): 0.0f}$ meV, 90% CL, "
                + rf"NME: {NME_PHENO_HIGH}- {NME_PHENO_LOW}",
            )

            plt.vlines(m_bb_pheno_low, 0, 1e-1, color=GOOD_ORANGE, alpha=0.8)
            plt.vlines(m_bb_pheno_high, 0, 1e-1, color=GOOD_ORANGE, alpha=0.8)

        # ------ get the legend-entries that are already attached to the axis
        colors, label = ax.get_legend_handles_labels()

        # ------ append the multicolor legend patches
        colors.append(
            MulticolorPatch(
                [
                    (BRAZIL_1_SIGMA_KWARGS["color"], BRAZIL_1_SIGMA_KWARGS["alpha"]),
                    (BRAZIL_2_SIGMA_KWARGS["color"], BRAZIL_2_SIGMA_KWARGS["alpha"]),
                ]
            )
        )
        label.append(r"median exp. bkg. only $\pm$ 1 $\sigma,\,\pm 2 \sigma$")

        if self.plot_nme_range:
            colors.append((band, obs))
            label.append(
                "obs., "
                + r"$m_{\beta\beta} <$ "
                + rf"${round(T_est*1000, -1): 0.0f}_{{-{round(T_est*1000-m_bb_belley_low, -1)}}}^{{+{round(m_bb_belley_high-T_est*1000, -1)}}}$ meV, 90% CL, "
                + rf"$\mathrm{{NME}}: {NME_central}_{{-{NME_unc_lo}}}^{{+{NME_unc_hi}}}$"
            )
            plt.axhline(y=0.1, ls=":", color="k")
            plt.vlines(T_est * 1000, 0, 1e-1, color="k")
            plt.vlines(T_12_median * 1000, 0, 1e-1, ls="--", color="k")
        else:
            colors.append(obs)
            label.append(
                "obs., "
                + r"$m_{\beta\beta} <$ "
                + rf"${round(T_est*1000, -1): 0.0f}$ meV, 90% CL"
            )

            plt.axhline(y=0.1, ls=":", **P_VALUE_KWARGS)
            plt.vlines(T_est * 1000, 0, 1e-1, **P_VALUE_KWARGS)
            plt.vlines(T_12_median * 1000, 0, 1e-1, ls="--", **P_VALUE_KWARGS)

        ax.xaxis.set_minor_locator(MultipleLocator(5))
        ax.minorticks_on()

        plt.yscale("log")
        plt.xlabel(r"$m_{\beta\beta}$  [meV]")
        plt.ylabel(r"$p$-value")

        # plt.legend()
        ax.legend(
            colors,
            label,
            loc="best",
            handler_map={MulticolorPatch: MulticolorPatchHandler()},
            frameon=False,
        )
        legendstyles.legend_watermark(ax, logo_suffix="-200", approved=True)
        # legendstyles.add_preliminary(ax, color="red")

        plt.ylim(*self.sensitivity_y_lim)
        plt.xlim(*self.nme_sensitivity_x_lim)
        plt.tight_layout()
        return fig
